"""Widget generators and their signal handlers"""

import os
from abc import ABC, abstractmethod
from gettext import gettext as _
from inspect import Parameter, signature
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

from gi.repository import Gdk, Gtk  # type: ignore

from lutris.config import LutrisConfig
from lutris.gui.widgets import NotificationSource
from lutris.gui.widgets.common import EditableGrid, FileChooserEntry, Label
from lutris.gui.widgets.searchable_entrybox import SearchableEntrybox
from lutris.util.log import logger
from lutris.util.strings import gtk_safe

if TYPE_CHECKING:
    from lutris.gui.config.boxes import ConfigBox


class WidgetGenerator(ABC):
    """This class generates widgets for an options page. It even adds the widgets
    to a 'wrapper' you can supply and, if required, that wrapper goes into a container
    which also contains message boxes for errors and warnings.

    The specific inner widgets are generated according to an option dict.

    The generator also accumulates and tracks the wrappers and containers for future use,
    and can update their state via a call to update_widgets(). When doing this, various callables
    in the option dicts can be re-evaluated; you provide a set of 'callback_args' and 'callback_kwargs'
    that are passed along to these callbacks. These collbacks also take the identifier of the
    option, from the 'option' key of the option dict.

    'changed' is a NotificationSource for whenever the generated widget changes a value;
    the same 'changed' NotificationSource is shared by all the widgets this generator generates.

    You can only create new containers, not discard old ones, but you *can* regenerate the widget
    inside the wrapper as a way to force it to update.
    """

    GeneratorFunction = Callable[[Dict[str, Any], Any, Any], Optional[Gtk.Widget]]

    def __init__(self, parent: "ConfigBox", *callback_args, **callback_kwargs) -> None:
        self.parent = parent
        self.callback_args = callback_args
        self.callback_kwargs = callback_kwargs
        self.changed = NotificationSource()  # takes option_key, new_value
        self.changed.register(self.on_changed, priority=1000)
        self._default_directory: Optional[str] = None
        self._current_parent: Optional[Gtk.Box] = None
        self._current_section: Optional[str] = None

        # These are outputs set by generate_widget() or generate_container()
        # and they are reset on each call.
        self.wrapper: Optional[Gtk.Box] = None
        self.default_value = None
        self.tooltip_default: Optional[str] = None
        self.options: Dict[str, Dict[str, Any]] = {}
        self.option_widget: Optional[Gtk.Widget] = None
        self.option_container: Optional[Gtk.Widget] = None
        self.warning_messages: List[Gtk.Widget] = []

        # These accumulate results across all widgets
        self.wrappers: Dict[str, Gtk.Container] = {}
        self.section_frames: List[SectionFrame] = []
        self.option_containers: Dict[str, Gtk.Container] = {}

        self._generators: Dict[str, WidgetGenerator.GeneratorFunction] = {
            "label": self._generate_label,
            "string": self._generate_string,
            "bool": self._generate_bool,
            "range": self._generate_range,
            "choice": self._generate_choice,
            "choice_with_entry": self._generate_choice_with_entry,
            "choice_with_search": self._generate_choice_with_search,
            "file": self._generate_file,
            "multiple_file": self._generate_multiple_file,
            "directory": self._generate_directory,
            "mapping": self._generate_mapping,
            "command_line": self._generate_command_line,
        }

    def on_changed(self, option_key: str, new_value: Any) -> None:
        """Called when any value is changed; this is called later than ordinary
        handlers for the 'changed' notification, and by default just updates the
        widgets."""
        self.update_widgets()

    @property
    def default_directory(self) -> str:
        """This is the directory selected by default by file and directory choosers."""
        if not self._default_directory:
            lutris_config = LutrisConfig()
            self._default_directory = lutris_config.system_config.get("game_path") or os.path.expanduser("~")
        return self._default_directory

    @default_directory.setter
    def default_directory(self, new_dir: str) -> None:
        self._default_directory = new_dir

    # Widget Construction

    def add_container(self, option: Dict[str, Any], wrapper: Optional[Gtk.Box] = None) -> Optional[Gtk.Widget]:
        """Generates the option's widget, wrapper and container, and adds the container to the parent;
        if the option uses 'section', then the container is actually placed inside a SectionFrame,
        or in the previous frame if it is for the same section."""
        option_container = self.generate_container(option, wrapper)

        if option_container and self.parent:
            # Switch to new section if required
            if not self._current_parent:
                self._current_parent = self.parent

            if option.get("section") != self._current_section:
                self._current_section = option.get("section")
                if self._current_section:
                    frame = SectionFrame(self._current_section, visible=True)
                    self.section_frames.append(frame)
                    self._current_parent = frame.vbox
                    self.parent.pack_start(frame, False, False, 0)
                else:
                    self._current_parent = self.parent

            self._current_parent.pack_start(option_container, False, False, 0)
        return option_container

    def generate_container(self, option: Dict[str, Any], wrapper: Optional[Gtk.Box] = None) -> Optional[Gtk.Widget]:
        """Creates the widget, wrapper, and container; this returns the container
        (or the wrapper if there's no container)."""
        option_widget = self.generate_widget(option, wrapper)
        if option_widget and self.wrapper:
            option_key = option["option"]
            option_container = self.create_option_container(option, self.wrapper)
            self.option_containers[option_key] = option_container
            option_container.show_all()

            option_container.lutris_option_key = option_key  # type:ignore[attr-defined]
            option_container.lutris_option_label = option["label"]  # type:ignore[attr-defined]
            option_container.lutris_option_helptext = option.get("help") or ""  # type:ignore[attr-defined]

            # Mark advanced option containers, to be hidden by checking for this
            option_container.lutris_advanced = bool(option.get("advanced"))  # type:ignore[attr-defined]
            option_container.lutris_option = option  # type:ignore[attr-defined]

            self.option_container = option_container
            return option_container
        else:
            return None

    def generate_widget(self, option: Dict[str, Any], wrapper: Optional[Gtk.Box] = None) -> Optional[Gtk.Widget]:
        """This creates a wrapper box and a label and widget within it according to the options dict
        given. The option widget itself, is returned, but this method also sets attributes on the
        generator. You get 'wrapper', 'default_value', 'tooltip_default' and 'option_widget' which restates
        the return value. This returns None if the entire option should be omitted."""
        option_key = option["option"]
        option_type = option["type"]
        default = self.get_default(option)
        value = self.get_setting(option_key, default)

        self.default_value = default
        self.tooltip_default = None
        self.option_widget = None
        self.option_container = None
        self.warning_messages.clear()
        self.wrappers.pop(option_key, None)

        # Record the options themselves before anything is generated
        self.options[option_key] = option

        if wrapper:
            # Destroy and recreate option widget
            children = wrapper.get_children()
            for child in children:
                child.destroy()
            self.wrapper = wrapper
        else:
            self.wrapper = self.create_wrapper_box(option, value, default)

            if not self.wrapper:
                return None

        func = self._generators.get(option_type)

        if func:
            option_widget = func(option, value, default)
        else:
            raise ValueError("Unknown widget type %s" % option_type)

        self.wrappers[option_key] = self.wrapper
        self.option_widget = option_widget
        self.tooltip_default = self.tooltip_default or (default if isinstance(default, str) else None)

        if option_widget:
            option_widget.show_all()

        self.configure_wrapper_box(self.wrapper, option, value, default)
        self.configure_warning_messages(option)
        return option_widget

    def configure_wrapper_box(self, wrapper: Gtk.Widget, option: Dict[str, Any], value: Any, default: Any) -> None:
        """Configures the wrapper box after it is created; this sets its tooltip, sensitivity, and
        creates warning message boxes."""

        # Attach a tooltip to the wrapper
        tooltip = self.get_tooltip(option, value, default)
        if tooltip:
            wrapper.props.has_tooltip = True
            wrapper.connect("query-tooltip", self.on_query_tooltip, tooltip)

    def get_tooltip(self, option: Dict[str, Any], value: Any, default: Any):
        tooltip = option.get("help")
        if self.tooltip_default:
            tooltip = tooltip + "\n\n" if tooltip else ""
            tooltip += _("<b>Default</b>: ") + self.tooltip_default
        return tooltip

    def configure_warning_messages(self, option: Dict[str, Any]):
        # Add message boxes under the widget
        if "error" in option:
            self.warning_messages.append(ConfigErrorBox(option["error"]))

        if "warning" in option:
            self.warning_messages.append(ConfigWarningBox(option["warning"]))

    def create_wrapper_box(self, option: Dict[str, Any], value: Any, default: Any) -> Optional[Gtk.Box]:
        """This creates the wrapper, which becomes the 'wrapper' attribute and which build_option_widget()
        populates. Returns None if the option is not visible; in that case no widget is generated either."""

        available = self._evaluate_flag_option("available", option)

        if not available:
            # If not available, there's no wrapper, and no widget!
            return None

        return Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12, margin_bottom=6, visible=True)

    def create_option_container(self, option: Dict[str, Any], wrapper: Gtk.Container) -> Gtk.Container:
        """This creates a wrapper box around the widget wrapper, to support additional controls. The
        base implementation wraps 'wrapper' in a Box with the error and warning widgets; if
        there are none it just returns 'wrapper'."""

        if self.warning_messages:
            option_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True)
            option_container.pack_start(wrapper, False, False, 0)

            for widget in self.warning_messages:
                option_container.pack_start(widget, False, False, 0)

            return option_container
        else:
            return wrapper

    def build_option_widget(
        self, option: Dict[str, Any], widget: Optional[Gtk.Widget], no_label: bool = False, expand: bool = True
    ) -> Optional[Gtk.Widget]:
        """This is called by the generator methods to place their widget into the wrapper, usually with
        a label taken from 'option'.

        The default labeling and placement is suitable for our ConfigBoxes, but we override this to
        get a different layout for PreferencesBox.

        Some generators do their own labelling, and pass True for no_label; this method will
        pack the widget with no label in that case.

        This method returns the widget."""
        if self.wrapper:
            if not no_label and "label" in option:
                label = option["label"]
                label = Label(label)
                self.wrapper.pack_start(label, False, False, 0)

            if widget:
                option_size = option.get("size", None)
                if option_size:
                    expand = option_size != "small"

                self.wrapper.pack_start(widget, expand, expand, 0)
        return widget

    # Dynamic Widget Updates

    def update_widgets(self) -> None:
        """Call this to update the visibility, sensitivity and other properties of
        the widgets, wrappers and containers already generated."""

        for option_key, container in self.option_containers.items():
            if hasattr(container, "lutris_option"):
                option = container.lutris_option
                wrapper = self.wrappers[option_key]
                self.update_option_container(option, container, wrapper)

        for frame in self.section_frames:
            visible = frame.has_visible_children()
            frame.set_visible(visible)
            frame.set_no_show_all(not visible)

    def update_option_container(self, option, container: Gtk.Container, wrapper: Gtk.Container):
        """This method updates an option container and its wrapper; this re-evaluates the
        relevant options in case they contain callables and those callables return different
        results."""

        # Update messages in message boxes that support it
        for child in container.get_children():
            if hasattr(child, "update_message"):
                child.update_message(option, self)

        # Hide entire container if the option is not visible
        visible = self.get_visibility(option)
        container.set_visible(visible)
        container.set_no_show_all(not visible)

        # Grey out option if condition unmet, or if a second setting is False
        condition: bool = self.get_condition(option)
        wrapper.set_sensitive(condition)

    # Widget factories

    # Label
    def _generate_label(self, option, value, default):
        """Generate a simple label."""
        text = option["label"]
        label = Label(text)
        label.set_use_markup(True)
        label.set_halign(Gtk.Align.START)
        label.set_valign(Gtk.Align.CENTER)
        return self.build_option_widget(option, label, no_label=True)

    # Entry
    def _generate_string(self, option, value, default):
        """Generate an entry box."""

        def on_changed(entry):
            """Action triggered for entry 'changed' signal."""
            self.changed.fire(option_key, entry.get_text())

        option_key = option["option"]

        entry = Gtk.Entry()
        entry.set_text(value or default or "")
        entry.connect("changed", on_changed)
        return self.build_option_widget(option, entry)

    # Switch
    def _generate_bool(self, option, value, default):
        """Generate a switch."""

        def on_notify_active(widget, _gparam):
            """Action for the switch's toggled signal."""
            self.changed.fire(option_key, widget.get_active())

        def to_bool(to_convert):
            """Convert values to booleans in a way that won't decide that
            the string 'False' is True!"""
            if to_convert is None:
                return None
            elif isinstance(to_convert, str):
                text = to_convert.casefold().strip()
                if text == "true":
                    return True
                elif text == "false":
                    return False
                else:
                    return None
            else:
                return bool(to_convert)

        option_key = option["option"]

        active = to_bool(value)
        if active is None:
            active = bool(to_bool(default))

        switch = Gtk.Switch(active=active, valign=Gtk.Align.CENTER)
        switch.connect("notify::active", on_notify_active)

        self.tooltip_default = _("Enabled") if default else _("Disabled")
        return self.build_option_widget(option, switch, expand=False)

    # SpinButton
    def _generate_range(self, option, value, default):
        """Generate a ranged spin button."""

        def on_changed(widget):
            """Action triggered on spin button 'changed' signal."""
            new_value = widget.get_value_as_int()
            self.changed.fire(option_key, new_value)

        option_key = option["option"]
        min_val = option["min"]
        max_val = option["max"]

        adjustment = Gtk.Adjustment(float(min_val), float(min_val), float(max_val), 1, 0, 0)
        spin_button = Gtk.SpinButton()
        spin_button.set_adjustment(adjustment)
        spin_button.set_value(value or default or 0)
        spin_button.connect("changed", on_changed)
        return self.build_option_widget(option, spin_button)

    # ComboBox
    def _generate_choice(self, option, value, default, has_entry=False):
        """Generate a combobox (drop-down menu)."""

        def populate_combobox_choices():
            expanded, tooltip_default, _valid_choices = expand_combobox_choices()
            for choice in expanded:
                liststore.append(choice)

            if tooltip_default:
                self.tooltip_default = str(tooltip_default)

        def expand_combobox_choices():
            expanded = []
            tooltip_default = None
            valid = []
            has_value = False
            for choice in choices:
                if isinstance(choice, str):
                    choice = (choice, choice)
                if choice[1] == value:
                    has_value = True
                if choice[1] == default:
                    tooltip_default = choice[0]
                    choice = (_("%s (default)") % choice[0], choice[1])
                valid.append(choice[1])
                expanded.append(choice)
            if not has_value and value:
                expanded.insert(0, (value, value))
            return expanded, tooltip_default, valid

        def on_combobox_scroll(widget, _event):
            """Prevents users from accidentally changing configuration values
            while scrolling down dialogs.
            """
            widget.stop_emission_by_name("scroll-event")
            return False

        def on_combobox_change(widget):
            """Action triggered on combobox 'changed' signal."""
            list_store = widget.get_model()
            active = widget.get_active()
            option_value = None
            if active < 0:
                if widget.get_has_entry():
                    option_value = widget.get_child().get_text()
            else:
                option_value = list_store[active][1]
            self.changed.fire(option_key, option_value)

        option_key = option["option"]
        choices = self._evaluate_option("choices", None, option)

        liststore = Gtk.ListStore(str, str)
        populate_combobox_choices()
        # With entry ("choice_with_entry" type)
        if has_entry:
            combobox = Gtk.ComboBox.new_with_model_and_entry(liststore)
            combobox.set_entry_text_column(0)
        # No entry ("choice" type)
        else:
            combobox = Gtk.ComboBox.new_with_model(liststore)
            cell = Gtk.CellRendererText()
            combobox.pack_start(cell, True)
            combobox.add_attribute(cell, "text", 0)

        combobox.set_id_column(1)

        expanded_choices, _tooltip_default, valid_choices = expand_combobox_choices()
        if value in [v for _k, v in expanded_choices]:
            combobox.set_active_id(value)
        elif has_entry:
            for ch in combobox.get_children():
                if isinstance(ch, Gtk.Entry):
                    ch.set_text(value or "")
                    break
        else:
            combobox.set_active_id(default)

        combobox.connect("changed", on_combobox_change)
        combobox.connect("scroll-event", on_combobox_scroll)
        combobox.set_valign(Gtk.Align.CENTER)

        def get_invalidity_error(key: str):
            v = self.get_setting(key, self.get_default(option))
            if v in valid_choices:
                return None

            return _("The setting '%s' is no longer available. You should select another choice.") % v

        if not has_entry and value not in valid_choices:
            self.warning_messages.append(ConfigWarningBox(get_invalidity_error))

        return self.build_option_widget(option, combobox)

    # ComboBox
    def _generate_choice_with_entry(self, option, value, default):
        return self._generate_choice(option, value, default, has_entry=True)

    # Searchable Entry
    def _generate_choice_with_search(self, option, value, default):
        """Generate a searchable combo box"""

        def on_changed(_widget, new_value):
            self.changed.fire(option_key, new_value)

        option_key = option["option"]
        choices = option["choices"]
        entrybox = SearchableEntrybox(choices, value or default)
        entrybox.connect("changed", on_changed)
        return self.build_option_widget(option, entrybox)

    # FileChooserEntry
    def _generate_file(self, option, value, default, shell_quoting=False):
        """Generate a file chooser button to select a file."""

        def on_changed(entry):
            """Action triggered when the field's content changes."""
            text = entry.get_text()
            self.changed.fire(option_key, text)

        option_key = option["option"]
        warn_if_non_writable_parent = bool(option.get("warn_if_non_writable_parent"))

        if not value:
            value = default

        if "default_path" in option:
            lutris_config = LutrisConfig()
            chooser_default_path = lutris_config.system_config.get(option["default_path"])
        else:
            chooser_default_path = self.default_directory

        file_chooser = FileChooserEntry(
            title=_("Select file"),
            action=Gtk.FileChooserAction.OPEN,
            warn_if_non_writable_parent=warn_if_non_writable_parent,
            text=value,
            default_path=chooser_default_path,
            shell_quoting=shell_quoting,
        )

        if value:
            # If path is relative, complete with default directory
            if not os.path.isabs(value):
                value = os.path.expanduser(value)
                if not os.path.isabs(value):
                    value = os.path.join(self.default_directory, value)
            file_chooser.entry.set_text(value)

        file_chooser.set_valign(Gtk.Align.CENTER)
        file_chooser.connect("changed", on_changed)

        return self.build_option_widget(option, file_chooser)

    # FileChooserEntry
    def _generate_command_line(self, option, value, default):
        return self._generate_file(option, value, default, shell_quoting=True)

    # TreeView
    def _generate_multiple_file(self, option, value, default):
        """Generate a multiple file selector."""

        def on_add_files_clicked(_widget):
            """Create and run multi-file chooser dialog."""

            dialog = Gtk.FileChooserNative.new(
                _("Select files"),
                None,
                Gtk.FileChooserAction.OPEN,
                _("_Add"),
                _("_Cancel"),
            )
            dialog.set_select_multiple(True)

            files = [row[0] for row in files_list_store]
            first_file_dir = os.path.dirname(files[0]) if files else None
            dialog.set_current_folder(first_file_dir or self.default_directory)
            response = dialog.run()
            if response == Gtk.ResponseType.ACCEPT:
                for filename in dialog.get_filenames():
                    if filename not in files:
                        files_list_store.append([filename])
                        files.append(filename)
                self.changed.fire(option_key, files)
            dialog.destroy()

        def on_files_treeview_keypress(treeview, event):
            """Action triggered when a row is deleted from the filechooser."""
            if event.keyval == Gdk.KEY_Delete:
                selection = treeview.get_selection()
                (model, treepaths) = selection.get_selected_rows()
                for treepath in treepaths:
                    treeiter = model.get_iter(treepath)
                    model.remove(treeiter)

                    files = [row[0] for row in files_list_store]
                    self.changed.fire(option_key, files)

        option_key = option["option"]
        label = option["label"]

        files_list_store = Gtk.ListStore(str)
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        label = Label(label + ":")
        label.set_halign(Gtk.Align.START)
        button = Gtk.Button(label=_("Add Files"))
        button.connect("clicked", on_add_files_clicked)
        button.set_margin_left(10)
        vbox.pack_start(label, False, False, 5)
        vbox.pack_end(button, False, False, 0)

        if not value:
            value = default

        if value:
            if isinstance(value, str):
                files = [value]
            else:
                files = value
        else:
            files = []
        for filename in files:
            files_list_store.append([filename])
        cell_renderer = Gtk.CellRendererText()
        files_treeview = Gtk.TreeView(model=files_list_store)
        files_column = Gtk.TreeViewColumn(_("Files"), cell_renderer, text=0)
        files_treeview.append_column(files_column)
        files_treeview.connect("key-press-event", on_files_treeview_keypress)
        treeview_scroll = Gtk.ScrolledWindow()
        treeview_scroll.set_min_content_height(130)
        treeview_scroll.set_margin_left(10)
        treeview_scroll.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        treeview_scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        treeview_scroll.add(files_treeview)

        vbox.pack_start(treeview_scroll, True, True, 0)
        return self.build_option_widget(option, vbox, no_label=True)

    # FileChooserEntry
    def _generate_directory(self, option, value, default):
        """Generate a file chooser button to select a directory."""

        def on_changed(entry):
            """Action triggered when the field's content changes."""
            text = entry.get_text()
            self.changed.fire(option_key, text)

        option_key = option["option"]
        warn_if_non_writable_parent = bool(option.get("warn_if_non_writable_parent"))

        if not value:
            value = default

        directory_chooser = FileChooserEntry(
            title=_("Select folder"),
            action=Gtk.FileChooserAction.SELECT_FOLDER,
            warn_if_non_writable_parent=warn_if_non_writable_parent,
            text=value,
            default_path=self.default_directory if not value else None,
        )
        directory_chooser.connect("changed", on_changed)
        directory_chooser.set_valign(Gtk.Align.CENTER)
        return self.build_option_widget(option, directory_chooser)

    # EditableGrid
    def _generate_mapping(self, option, value, default):
        """Adds an editable grid widget"""

        def on_changed(widget):
            values = dict(widget.get_data())
            self.changed.fire(option_key, values)

        option_key = option["option"]
        value = value or default or {}
        try:
            value = list(value.items())
        except AttributeError:
            logger.error("Invalid value of type %s passed to grid widget: %s", type(value), value)
            value = {}

        grid = EditableGrid(value, columns=["Key", "Value"])
        grid.connect("changed", on_changed)
        return self.build_option_widget(option, grid)

    @staticmethod
    def on_query_tooltip(_widget, _x, _y, _keybmode, tooltip, text):  # pylint: disable=unused-argument
        """Prepare a custom tooltip with a fixed width"""
        label = Label(text)
        label.set_use_markup(True)
        label.set_max_width_chars(60)
        event_box = Gtk.EventBox()
        event_box.add(label)
        event_box.show_all()
        tooltip.set_custom(event_box)
        return True

    # Option access

    @abstractmethod
    def get_setting(self, option_key: str, default: Any) -> Any:
        """Reads the current value for a specific setting; this method must be
        implemented by a subclass."""
        raise NotImplementedError()

    def get_default(self, option: Dict[str, Any]) -> Any:
        """Returns the default value from the option; if it is callable, this calls
        it to get the actual default."""
        return self._evaluate_option("default", default=None, option=option)

    def get_visibility(self, option: Dict[str, Any]) -> bool:
        """Extracts the 'visible' option; if the option is missing this returns
        True, and if it is callable this calls it. Subclasses can add further conditions."""
        return self._evaluate_flag_option("visible", option)

    def get_condition(self, option: Dict[str, Any]) -> bool:
        """Extracts the 'condition' option; but also the 'conditional_on' option, and if both
        are present, then if either indicates the control should be disabled this will be false.."""
        condition = self._evaluate_flag_option("condition", option)
        conditional_on = option.get("conditional_on")

        if conditional_on:
            conditional_on_default = self.get_default(self.options[conditional_on])
            if not self.get_setting(conditional_on, conditional_on_default):
                return False

        container = self.option_containers[option["option"]]

        for child in container.get_children():
            if hasattr(child, "blocks_sensitivity") and child.blocks_sensitivity:
                return False

        return condition

    def _evaluate_flag_option(self, key: str, option: Dict[str, Any]) -> bool:
        """Evaluates a flag option; if is None or missing this returns True, and if
        it is callable this calls it (as with _evaluate_option) and converts
        the result to a bool."""
        flag = self._evaluate_option(key, default=True, option=option)
        return bool(flag) if flag is not None else True

    def _evaluate_option(self, key: str, default: Any, option: Dict[str, Any]) -> Any:
        """Evaluates an option; if is missing, then function returns 'default', and
        if it is callable this calls it, passing the option key, generator's args and kwargs.

        The callable may take fewer arguments; if so, this will pass as many argments
        as it will take, even if that is none at all."""

        if key not in option:
            return default

        value = option[key]
        return self.evaluate_option_value(value, option=option)

    def evaluate_option_value(self, value: Any, option: Dict[str, Any]) -> Any:
        """Evaluates the 'value' given, if it is callable. If not, this method just
        returns the 'value'.

        The 'value' is called with the option-key and then all the callback arguments
        given to this generator's __init__. If the 'value' takes fewer arguments than
        this, trailing arguments are omitted."""
        if callable(value):
            sig = signature(value)
            argcount = len(sig.parameters)

            option_key = option["option"]
            argsneeded = 1 + len(self.callback_args)

            if argcount >= argsneeded:  # enough declared args?
                return value(option_key, *self.callback_args, **self.callback_kwargs)
            elif any(p.kind == Parameter.VAR_POSITIONAL for p in sig.parameters.values()):  # unlimited args via *args?
                return value(option_key, *self.callback_args, **self.callback_kwargs)
            elif argcount == 0:  # no args?
                return value(**self.callback_kwargs)
            else:  # any other number of args
                args = list(self.callback_args)
                args.insert(0, option_key)
                args = args[:argcount]
                return value(*args, **self.callback_kwargs)

        return value


class SectionFrame(Gtk.Frame):
    """A frame that is styled to have particular margins, and can have its frame hidden.
    This leaves the content but removes the margins and borders and all that, so it looks
    like the frame was never there."""

    def __init__(self, section, **kwargs):
        super().__init__(label=section, **kwargs)
        self.section = section
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True)
        self.add(self.vbox)
        self.get_style_context().add_class("section-frame")

    def has_visible_children(self):
        return any(w for w in self.vbox.get_children() if w.get_visible())


class WidgetWarningMessageBox(Gtk.Box):
    """A box to display a message with an icon inside the configuration dialog."""

    def __init__(self, icon_name, margin_left=18, margin_right=18, margin_bottom=6):
        super().__init__(
            spacing=6,
            visible=False,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            no_show_all=True,
        )

        image = Gtk.Image(visible=True)
        image.set_from_icon_name(icon_name, Gtk.IconSize.DND)
        self.pack_start(image, False, False, 0)
        self.label = Gtk.Label(visible=True, xalign=0)
        self.label.set_line_wrap(True)
        self.pack_start(self.label, False, False, 0)

    def show_markup(self, markup) -> bool:
        """Displays the markup given, and shows this box. If markup is empty or None,
        this hides the box instead. Returns the new visibility."""
        visible = bool(markup)

        if markup:
            self.label.set_markup(str(markup))

        self.set_visible(visible)
        return visible


class ConfigMessageBox(WidgetWarningMessageBox):
    def __init__(self, message, icon_name, **kwargs):
        self.message = message
        super().__init__(icon_name, **kwargs)

        if not callable(message):
            text = gtk_safe(message)

            if text:
                self.label.set_markup(str(text))

    def update_message(self, option: Dict[str, Any], generator: WidgetGenerator) -> bool:
        try:
            text = generator.evaluate_option_value(self.message, option)
        except Exception as err:
            logger.exception("Unable to generate configuration warning: %s", err)
            text = gtk_safe(str(err))

        return self.show_markup(text)


class ConfigWarningBox(ConfigMessageBox):
    def __init__(self, warning):
        super().__init__(warning, icon_name="dialog-warning")


class ConfigErrorBox(ConfigMessageBox):
    def __init__(self, error):
        super().__init__(error, icon_name="dialog-error")

    @property
    def blocks_sensitivity(self):
        """Called to check if the wrapper should be made insensitive
        because of this error box."""
        return self.get_visible()



"""Runner for MAME"""

import os
from gettext import gettext as _

from lutris import runtime, settings
from lutris.exceptions import GameConfigError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger
from lutris.util.mame.database import get_supported_systems
from lutris.util.strings import split_arguments

MAME_CACHE_DIR = os.path.join(settings.CACHE_DIR, "mame")
MAME_XML_PATH = os.path.join(MAME_CACHE_DIR, "mame.xml")


def write_mame_xml(force=False):
    if not system.path_exists(MAME_CACHE_DIR):
        system.create_folder(MAME_CACHE_DIR)
    if system.path_exists(MAME_XML_PATH, exclude_empty=True) and not force:
        return False
    logger.info("Writing full game list from MAME to %s", MAME_XML_PATH)
    mame_inst = mame()
    mame_inst.write_xml_list()
    if system.get_disk_size(MAME_XML_PATH) == 0:
        logger.warning("MAME did not write anything to %s", MAME_XML_PATH)
        return False
    return True


def notify_mame_xml(result, error):
    if error:
        logger.error("Failed writing MAME XML")
    elif result:
        logger.info("Finished writing MAME XML")


def get_system_choices(include_year=True):
    """Return list of systems for inclusion in dropdown"""
    if not system.path_exists(MAME_XML_PATH, exclude_empty=True):
        mame_inst = mame()
        if mame_inst.is_installed():
            AsyncCall(write_mame_xml, notify_mame_xml)
        return []
    for system_id, info in sorted(
        get_supported_systems(MAME_XML_PATH).items(),
        key=lambda sys: (sys[1]["manufacturer"], sys[1]["description"]),
    ):
        if info["description"].startswith(info["manufacturer"]):
            template = ""
        else:
            template = "%(manufacturer)s "
        template += "%(description)s"
        if include_year:
            template += " %(year)s"
        system_name = template % info
        system_name = system_name.replace("<generic>", "").strip()
        yield (system_name, system_id)


class mame(Runner):  # pylint: disable=invalid-name
    """MAME runner"""

    human_name = _("MAME")
    description = _("Arcade game emulator")
    runner_executable = "mame/mame"
    flatpak_id = "org.mamedev.MAME"
    runnable_alone = True
    config_dir = os.path.expanduser("~/.mame")
    cache_dir = os.path.join(settings.CACHE_DIR, "mame")
    xml_path = os.path.join(cache_dir, "mame.xml")
    _platforms = []

    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
        },
        {
            "option": "machine",
            "type": "choice_with_search",
            "label": _("Machine"),
            "choices": get_system_choices,
            "help": _("The emulated machine."),
        },
        {
            "option": "device",
            "type": "choice_with_entry",
            "label": _("Storage type"),
            "choices": [
                (_("Floppy disk"), "flop"),
                (_("Floppy drive 1"), "flop1"),
                (_("Floppy drive 2"), "flop2"),
                (_("Floppy drive 3"), "flop3"),
                (_("Floppy drive 4"), "flop4"),
                (_("Cassette (tape)"), "cass"),
                (_("Cassette 1 (tape)"), "cass1"),
                (_("Cassette 2 (tape)"), "cass2"),
                (_("Cartridge"), "cart"),
                (_("Cartridge 1"), "cart1"),
                (_("Cartridge 2"), "cart2"),
                (_("Cartridge 3"), "cart3"),
                (_("Cartridge 4"), "cart4"),
                (_("Snapshot"), "snapshot"),
                (_("Hard Disk"), "hard"),
                (_("Hard Disk 1"), "hard1"),
                (_("Hard Disk 2"), "hard2"),
                (_("CD-ROM"), "cdrm"),
                (_("CD-ROM 1"), "cdrm1"),
                (_("CD-ROM 2"), "cdrm2"),
                (_("Snapshot (dump)"), "dump"),
                (_("Quickload"), "quickload"),
                (_("Memory Card"), "memc"),
                (_("Cylinder"), "cyln"),
                (_("Punch Tape 1"), "ptap1"),
                (_("Punch Tape 2"), "ptap2"),
                (_("Print Out"), "prin"),
                (_("romimage1"), "rom1"),
                (_("romimage2"), "rom2"),
                (_("romimage3"), "rom3"),
                (_("romimage4"), "rom4"),
                (_("romimage5"), "rom5"),
                (_("midiin"), "min"),
                (_("midiout"), "mout"),
                (_("bitbanger"), "bitb"),
                (_("microtape1"), "utap1"),
                (_("microtape2"), "utap2"),
                (_("magtape1"), "mtap1"),
                (_("magtape2"), "mtap2"),
                (_("magtape3"), "mtap3"),
                (_("magtape4"), "mtap4"),
            ],
        },
        {
            "option": "args",
            "type": "string",
            "label": _("Arguments"),
            "help": _("Command line arguments used when launching the game"),
        },
        {
            "option": "slots",
            "type": "string",
            "label": ("Slot System"),
            "help": ("For slot devices that is needed for romsloads"),
        },
        {
            "option": "autoboot_command",
            "type": "string",
            "section": _("Autoboot"),
            "label": _("Autoboot command"),
            "help": _("Autotype this command when the system has started, an enter keypress is automatically added."),
        },
        {
            "option": "autoboot_delay",
            "type": "range",
            "section": _("Autoboot"),
            "label": _("Delay before entering autoboot command"),
            "min": 0,
            "max": 120,
        },
    ]

    runner_options = [
        {
            "option": "rompath",
            "type": "directory",
            "label": _("ROM/BIOS path"),
            "help": _(
                "Choose the folder containing ROMs and BIOS files.\n"
                "These files contain code from the original hardware "
                "necessary to the emulation."
            ),
        },
        {
            "option": "fullscreen",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Fullscreen"),
            "default": True,
        },
        {
            "option": "crt",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("CRT effect ()"),
            "help": _("Applies a CRT effect to the screen.Requires OpenGL renderer."),
            "default": False,
        },
        {
            "option": "verbose",
            "type": "bool",
            "section": _("Debugging"),
            "label": _("Verbose"),
            "help": _("display additional diagnostic information."),
            "default": False,
            "advanced": True,
        },
        {
            "option": "video",
            "type": "choice",
            "section": _("Graphics"),
            "label": _("Video backend"),
            "choices": (
                (_("Auto"), ""),
                ("OpenGL", "opengl"),
                ("BGFX", "bgfx"),
                ("SDL2", "accel"),
                (_("Software"), "soft"),
            ),
            "default": "opengl",
        },
        {
            "option": "waitvsync",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Wait for VSync"),
            "help": _("Enable waiting for  the  start  of  vblank  before flipping  screens; reduces tearing effects."),
            "advanced": True,
            "default": False,
        },
        {
            "option": "uimodekey",
            "type": "choice_with_entry",
            "label": _("Menu mode key"),
            "choices": [
                (_("Scroll Lock"), "SCRLOCK"),
                (_("Num Lock"), "NUMLOCK"),
                (_("Caps Lock"), "CAPSLOCK"),
                (_("Menu"), "MENU"),
                (_("Right Control"), "RCONTROL"),
                (_("Left Control"), "LCONTROL"),
                (_("Right Alt"), "RALT"),
                (_("Left Alt"), "LALT"),
                (_("Right Super"), "RWIN"),
                (_("Left Super"), "LWIN"),
            ],
            "default": "SCRLOCK",
            "advanced": True,
            "help": _("Key to switch between Full Keyboard Mode and Partial Keyboard Mode (default: Scroll Lock)"),
        },
    ]

    @property
    def working_dir(self):
        return os.path.join(settings.RUNNER_DIR, "mame")

    @property
    def platforms(self):
        if self._platforms:
            return self.platforms
        self._platforms = [choice[0] for choice in get_system_choices(include_year=False)]
        self._platforms += [_("Arcade"), _("Nintendo Game & Watch")]
        return self._platforms

    def install(self, install_ui_delegate, version=None, callback=None):
        def on_runner_installed(*args):
            def on_mame_ready(result, error):
                notify_mame_xml(result, error)
                if callback:
                    callback(*args)

            AsyncCall(write_mame_xml, on_mame_ready)

        super().install(install_ui_delegate, version=version, callback=on_runner_installed)

    @property
    def default_path(self):
        """Return the default path, use the runner's rompath"""
        main_file = self.game_config.get("main_file")
        if main_file:
            return os.path.dirname(main_file)
        return self.runner_config.get("rompath")

    def write_xml_list(self):
        """Write the full game list in XML to disk"""
        env = runtime.get_env()
        listxml_command = self.get_command() + ["-listxml"]
        os.makedirs(self.cache_dir, exist_ok=True)
        output, error_output = system.execute_with_error(listxml_command, env=env)
        if output:
            with open(self.xml_path, "w", encoding="utf-8") as xml_file:
                xml_file.write(output)
            logger.info("MAME XML list written to %s", self.xml_path)
        else:
            logger.warning("Couldn't get any output for mame -listxml: %s", error_output)

    def get_platform(self):
        selected_platform = self.game_config.get("platform")
        if selected_platform:
            return self.platforms[int(selected_platform)]
        if self.game_config.get("machine"):
            machine_mapping = {choice[1]: choice[0] for choice in get_system_choices(include_year=False)}
            # get_system_choices() can return [] if not yet ready, so we'll return
            # None in that case.
            return machine_mapping.get(self.game_config["machine"])
        rom_file = os.path.basename(self.game_config.get("main_file", ""))
        if rom_file.startswith("gnw_"):
            return _("Nintendo Game & Watch")
        return _("Arcade")

    def prelaunch(self):
        if not system.path_exists(os.path.join(self.config_dir, "mame.ini")):
            try:
                os.makedirs(self.config_dir)
            except OSError:
                pass
            system.execute(
                self.get_command() + ["-createconfig", "-inipath", self.config_dir],
                env=runtime.get_env(),
                cwd=self.working_dir,
            )

    def get_shader_params(self, shader_dir, shaders):
        """Returns a list of CLI parameters to apply a list of shaders"""
        params = []
        shader_path = os.path.join(self.working_dir, "shaders", shader_dir)
        for index, shader in enumerate(shaders):
            params += ["-gl_glsl", "-glsl_shader_mame%s" % index, os.path.join(shader_path, shader)]
        return params

    def play(self):
        command = self.get_command() + ["-skip_gameinfo", "-inipath", self.config_dir]
        if self.runner_config.get("video"):
            command += ["-video", self.runner_config["video"]]
        if not self.runner_config.get("fullscreen"):
            command.append("-window")
        if self.runner_config.get("waitvsync"):
            command.append("-waitvsync")
        if self.runner_config.get("uimodekey"):
            command += ["-uimodekey", self.runner_config["uimodekey"]]

        if self.runner_config.get("crt"):
            command += self.get_shader_params("CRT-geom", ["Gaussx", "Gaussy", "CRT-geom-halation"])
            command += ["-nounevenstretch"]

        if self.runner_config.get("verbose"):
            command += ["-verbose", "-oslog", "-log"]

        if self.game_config.get("machine"):
            rompath = self.runner_config.get("rompath")
            if rompath:
                command += ["-rompath", rompath]
            command.append(self.game_config["machine"])
            for slot_arg in split_arguments(self.game_config.get("slots")):
                command.append(slot_arg)
            device = self.game_config.get("device")
            if not device:
                raise GameConfigError(_("No device is set for machine %s") % self.game_config["machine"])
            rom = self.game_config.get("main_file")
            if rom:
                command += ["-" + device, rom]
        else:
            rompath = os.path.dirname(self.game_config.get("main_file"))
            if not rompath:
                rompath = self.runner_config.get("rompath")
            rom = os.path.basename(self.game_config.get("main_file"))
            if not rompath:
                raise GameConfigError(_("The path '%s' is not set. please set it in the options.") % "rompath")
            command += ["-rompath", rompath, rom]

        if self.game_config.get("autoboot_command"):
            command += ["-autoboot_command", self.game_config["autoboot_command"] + "\\n"]
            if self.game_config.get("autoboot_delay"):
                command += ["-autoboot_delay", str(self.game_config["autoboot_delay"])]

        for arg in split_arguments(self.game_config.get("args")):
            command.append(arg)

        return {"command": command}



"""Handle game specific actions"""

# Standard Library
# pylint: disable=too-many-public-methods
import os
from gettext import gettext as _
from pathlib import Path
from typing import List

from gi.repository import Gio, Gtk

from lutris.config import duplicate_game_config
from lutris.database import games
from lutris.database.games import add_game, get_game_by_field
from lutris.game import Game
from lutris.gui import dialogs
from lutris.gui.config.add_game_dialog import AddGameDialog
from lutris.gui.config.edit_game import EditGameConfigDialog
from lutris.gui.config.edit_game_categories import EditGameCategoriesDialog
from lutris.gui.dialogs import InputDialog
from lutris.gui.dialogs.delegates import LaunchUIDelegate
from lutris.gui.dialogs.log import LogWindow
from lutris.gui.dialogs.uninstall_dialog import UninstallDialog
from lutris.gui.widgets.utils import open_uri
from lutris.monitored_command import MonitoredCommand
from lutris.services.lutris import download_lutris_media
from lutris.util import xdgshortcuts
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger
from lutris.util.standalone_scripts import generate_script
from lutris.util.steam import shortcut as steam_shortcut
from lutris.util.strings import gtk_safe, slugify
from lutris.util.system import path_exists


class GameActions:
    """These classes provide a set of action to apply to a game or list of games, and can be used
    to populate menus. The base class handles the no-games case, for which there are no actions. But
    it also includes the code for actions that are shared between the subclasses. It also has methods for
    actions that are invokes externally by the GameBar."""

    def __init__(self, window: Gtk.Window, application=None):
        self.application = application or Gio.Application.get_default()
        self.window = window  # also used as a LaunchUIDelegate and InstallUIDelegate

    def get_games(self):
        """Return the list of games that the actions apply to."""
        return []

    def get_game_actions(self):
        """Return a list of game actions and their callbacks, Each item is a tuple
        of two strs and a callable, the action ID, it's human-readable name, and
        a callback to invoke to perform it. Menu separators are represented hre
        as (None, "-", None).
        """
        return []

    def get_displayed_entries(self):
        """Return a dictionary of flags indicating which actions are visible; the keys
        are the action ids from get_game_actions(), and the values are booleans indicating
        the action's visibility."""
        return {}

    @property
    def is_game_launchable(self):
        for game in self.get_games():
            if game.is_installed and not self.is_game_running:
                return True

        return False

    def on_game_launch(self, *_args):
        """Launch a game"""

    @property
    def is_game_running(self):
        for game in self.get_games():
            if game.is_db_stored and self.application.is_game_running_by_id(game.id):
                return True
        return False

    def on_game_stop(self, *_args):
        """Stops the game"""
        games = self.get_running_games()
        for game in games:
            game.force_stop()

    def get_running_games(self):
        running_games = []
        for game in self.get_games():
            if game and game.is_db_stored:
                ids = self.application.get_running_game_ids()
                for game_id in ids:
                    if str(game_id) == game.id:
                        running_games.append(game)
        return running_games

    @property
    def is_installable(self):
        for game in self.get_games():
            if not game.is_installed:
                return True

        return False

    def on_install_clicked(self, *_args):
        """Install a game"""
        # Install the currently selected game in the UI
        for game in self.get_games():
            if not game.slug:
                game_id = game.id if game.is_db_stored else game.name
                raise RuntimeError("No game to install: %s" % game_id)
            game.install(launch_ui_delegate=self.window)

    def on_add_favorite_game(self, _widget):
        """Add to favorite Games list"""
        for game in self.get_games():
            game.mark_as_favorite(True)

    def on_delete_favorite_game(self, _widget):
        """delete from favorites"""
        for game in self.get_games():
            game.mark_as_favorite(False)

    def on_hide_game(self, _widget):
        """Add a game to the list of hidden games"""
        for game in self.get_games():
            game.mark_as_hidden(True)

    def on_unhide_game(self, _widget):
        """Removes a game from the list of hidden games"""
        for game in self.get_games():
            game.mark_as_hidden(False)

    def on_locate_installed_game(self, *_args):
        """Show the user a dialog to import an existing install to a DRM free service

        Params:
            games ([Game]): List of Game instances without a database ID, populated with fields the service can provides
        """
        for game in self.get_games():
            AddGameDialog(self.window, game=game, runner=game.runner_name)

    def on_view_game(self, _widget):
        """Callback to open a game on lutris.net"""
        for game in self.get_games():
            open_uri("https://lutris.net/games/%s" % game.slug.replace("_", "-"))

    @property
    def is_game_removable(self):
        for game in self.get_games():
            if game.is_installed or game.is_db_stored:
                return True

        return False

    def on_remove_game(self, *_args):
        """Callback that present the uninstall dialog to the user"""
        game_ids = [g.id for g in self.get_games() if g.is_installed or g.is_db_stored]
        application = Gio.Application.get_default()
        dlg = application.show_window(UninstallDialog, parent=self.window)
        dlg.add_games(game_ids)

    def on_edit_game_categories(self, _widget):
        """Edit game categories"""
        games = self.get_games()
        if len(games) == 1:
            # Individual games get individual separate windows
            self.application.show_window(EditGameCategoriesDialog, game=games[0], parent=self.window)
        else:

            def add_games(window):
                window.add_games(self.get_games())

            # Multi-select means a common categories window for all of them; we can wind
            # up adding games to it if it's already open
            self.application.show_window(EditGameCategoriesDialog, update_function=add_games, parent=self.window)


class MultiGameActions(GameActions):
    """This actions class handles actions on multiple games together, and only iof they
    are 'db stored' games, not service games. This supports a subset of the actions
    of SingleGameActions."""

    def __init__(self, games: List[Game], window: Gtk.Window, application=None):
        super().__init__(window, application)
        self.games = games

    def get_games(self):
        return self.games

    def get_game_actions(self):
        return [
            ("stop", _("Stop"), self.on_game_stop),
            (None, "-", None),
            ("category", _("Categories"), self.on_edit_game_categories),
            ("favorite", _("Add to favorites"), self.on_add_favorite_game),
            ("deletefavorite", _("Remove from favorites"), self.on_delete_favorite_game),
            ("hide", _("Hide game from library"), self.on_hide_game),
            ("unhide", _("Unhide game from library"), self.on_unhide_game),
            (None, "-", None),
            ("remove", _("Remove"), self.on_remove_game),
        ]

    def get_displayed_entries(self):
        return {
            "stop": self.is_game_running,
            "category": True,
            "favorite": any(g for g in self.games if not g.is_favorite),
            "deletefavorite": any(g for g in self.games if g.is_favorite),
            "hide": any(g for g in self.games if g.is_installed and not g.is_hidden),
            "unhide": any(g for g in self.games if g.is_hidden),
            "remove": self.is_game_removable,
        }


class SingleGameActions(GameActions):
    """This actions class handles actions on a single game, which is a 'db stored' game,
    not a service game. This provides the largest selection of actions, including many
    that are unique to it."""

    def __init__(self, game: Game, window: Gtk.Window, application=None):
        super().__init__(window, application)
        self.game = game

    def get_games(self):
        return [self.game]

    def get_game_actions(self):
        return [
            ("play", _("Play"), self.on_game_launch),
            ("stop", _("Stop"), self.on_game_stop),
            ("execute-script", _("Execute script"), self.on_execute_script_clicked),
            ("show_logs", _("Show logs"), self.on_show_logs),
            (None, "-", None),
            ("configure", _("Configure"), self.on_edit_game_configuration),
            ("category", _("Categories"), self.on_edit_game_categories),
            ("browse", _("Browse files"), self.on_browse_files),
            ("favorite", _("Add to favorites"), self.on_add_favorite_game),
            ("deletefavorite", _("Remove from favorites"), self.on_delete_favorite_game),
            ("hide", _("Hide game from library"), self.on_hide_game),
            ("unhide", _("Unhide game from library"), self.on_unhide_game),
            (None, "-", None),
            ("install", _("Install"), self.on_install_clicked),
            ("install_more", _("Install another version"), self.on_install_clicked),
            ("install_dlcs", _("Install DLCs"), self.on_install_dlc_clicked),
            ("update", _("Install updates"), self.on_update_clicked),
            ("add", _("Locate installed game"), self.on_locate_installed_game),
            ("desktop-shortcut", _("Create desktop shortcut"), self.on_create_desktop_shortcut),
            ("rm-desktop-shortcut", _("Delete desktop shortcut"), self.on_remove_desktop_shortcut),
            ("menu-shortcut", _("Create application menu shortcut"), self.on_create_menu_shortcut),
            ("rm-menu-shortcut", _("Delete application menu shortcut"), self.on_remove_menu_shortcut),
            ("steam-shortcut", _("Create Steam shortcut"), self.on_create_steam_shortcut),
            (
                "steam-bigpicture-shortcut",
                _("Create Steam Big-Picture shortcut"),
                self.on_create_steam_bigpicture_shortcut,
            ),
            ("rm-steam-shortcut", _("Delete Steam shortcut"), self.on_remove_steam_shortcut),
            ("view", _("View on Lutris.net"), self.on_view_game),
            ("duplicate", _("Duplicate"), self.on_game_duplicate),
            (None, "-", None),
            ("remove", _("Remove"), self.on_remove_game),
        ]

    def get_displayed_entries(self):
        """Return a dictionary of actions that should be shown for a game"""

        game = self.game
        has_steam = steam_shortcut.vdf_file_exists()
        if has_steam:
            has_steam_shortcut = steam_shortcut.shortcut_exists(game)
            is_steam_game = steam_shortcut.is_steam_game(game)
        else:
            has_steam_shortcut = False
            is_steam_game = False

        return {
            "duplicate": game.is_installed,
            "install": self.is_installable,
            "add": not game.is_installed,
            "play": self.is_game_launchable,
            "update": game.is_updatable,
            "install_dlcs": game.is_updatable,
            "stop": self.is_game_running,
            "configure": bool(game.is_installed),
            "browse": game.is_installed and game.runner_name != "browser",
            "show_logs": game.is_installed,
            "category": True,
            "favorite": not game.is_favorite,
            "deletefavorite": game.is_favorite,
            "install_more": not game.service and game.is_installed,
            "execute-script": bool(
                game.is_installed and game.has_runner and game.runner.system_config.get("manual_command")
            ),
            "desktop-shortcut": bool(
                game.is_installed and not xdgshortcuts.desktop_launcher_exists(game.slug, game.id)
            ),
            "menu-shortcut": bool(game.is_installed and not xdgshortcuts.menu_launcher_exists(game.slug, game.id)),
            "steam-shortcut": bool(has_steam and game.is_installed and not has_steam_shortcut and not is_steam_game),
            "rm-desktop-shortcut": bool(game.is_installed and xdgshortcuts.desktop_launcher_exists(game.slug, game.id)),
            "rm-menu-shortcut": bool(game.is_installed and xdgshortcuts.menu_launcher_exists(game.slug, game.id)),
            "rm-steam-shortcut": bool(game.is_installed and has_steam_shortcut and not is_steam_game),
            "remove": self.is_game_removable,
            "view": True,
            "hide": game.is_installed and not game.is_hidden,
            "unhide": game.is_hidden,
        }

    def on_game_launch(self, *_args):
        """Launch a game"""
        game = self.game
        if game.is_installed and game.is_db_stored:
            if not self.application.is_game_running_by_id(game.id):
                game.launch(launch_ui_delegate=self.window)

    def on_execute_script_clicked(self, _widget):
        """Execute the game's associated script"""
        game = self.game
        manual_command = game.runner.system_config.get("manual_command")
        if path_exists(manual_command):
            runner = game.runner
            env = runner.get_env()
            MonitoredCommand(
                [manual_command], include_processes=[os.path.basename(manual_command)], cwd=game.directory, env=env
            ).start()
            logger.info("Running %s in the background", manual_command)

    def on_show_logs(self, _widget):
        """Display game log"""
        game = self.game
        _buffer = game.log_buffer
        if not _buffer:
            logger.info("No log for game %s", game)
        return LogWindow(game=game, buffer=_buffer, application=self.application)

    def on_edit_game_configuration(self, _widget):
        """Edit game preferences"""
        self.application.show_window(EditGameConfigDialog, game=self.game, parent=self.window)

    def on_browse_files(self, _widget):
        """Callback to open a game folder in the file browser"""
        path = self.game.get_browse_dir()
        if not path:
            dialogs.NoticeDialog(_("This game has no installation directory"))
        elif path_exists(path):
            open_uri("file://%s" % path)
        else:
            dialogs.NoticeDialog(_("Can't open %s \nThe folder doesn't exist.") % path)

    def on_install_dlc_clicked(self, _widget):
        self.game.install_dlc(install_ui_delegate=self.window)

    def on_update_clicked(self, _widget):
        self.game.install_updates(install_ui_delegate=self.window)

    def on_create_menu_shortcut(self, *_args):
        """Add the selected game to the system's Games menu."""
        game = self.game
        launch_config_name = self._select_game_launch_config_name(game)
        if launch_config_name is not None:
            xdgshortcuts.create_launcher(game.slug, game.id, game.name, launch_config_name, menu=True)

    def on_create_steam_shortcut(self, *_args):
        """Add the selected game to steam as a nonsteam-game."""
        game = self.game
        launch_config_name = self._select_game_launch_config_name(game)
        if launch_config_name is not None:
            steam_shortcut.create_shortcut(game, launch_config_name)

    def on_create_steam_bigpicture_shortcut(self, *_args):
        """Add the selected game to steam as a nonsteam-game."""
        game = self.game
        launch_config_name = self._select_game_launch_config_name(game)

        if launch_config_name is not None:
            steam_shortcut.create_shortcut(game, launch_config_name, True)
            gameDict = games.get_game_by_field(game.id, "id")
            gamepath = f"{Path.home()!s}/.local/share/applications/lutris-{game.slug}.sh"
            generate_script(logger, LaunchUIDelegate(), gameDict, gamepath)

    def on_create_desktop_shortcut(self, *_args):
        """Create a desktop launcher for the selected game."""
        game = self.game
        launch_config_name = self._select_game_launch_config_name(game)
        if launch_config_name is not None:
            xdgshortcuts.create_launcher(game.slug, game.id, game.name, launch_config_name, desktop=True)

    def on_remove_menu_shortcut(self, *_args):
        """Remove an XDG menu shortcut"""
        game = self.game
        xdgshortcuts.remove_launcher(game.slug, game.id, menu=True)

    def on_remove_steam_shortcut(self, *_args):
        """Remove the selected game from list of non-steam apps."""
        steam_shortcut.remove_shortcut(self.game)

    def on_remove_desktop_shortcut(self, *_args):
        """Remove a .desktop shortcut"""
        game = self.game
        xdgshortcuts.remove_launcher(game.slug, game.id, desktop=True)

    def on_game_duplicate(self, _widget):
        game = self.game

        duplicate_game_dialog = InputDialog(
            {
                "parent": self.window,
                "question": _(
                    "Do you wish to duplicate %s?\nThe configuration will be duplicated, "
                    "but the games files will <b>not be duplicated</b>.\n"
                    "Please enter the new name for the copy:"
                )
                % gtk_safe(game.name),
                "title": _("Duplicate game?"),
                "initial_value": game.name,
            }
        )
        result = duplicate_game_dialog.run()
        if result != Gtk.ResponseType.OK:
            duplicate_game_dialog.destroy()
            return
        new_name = duplicate_game_dialog.user_value

        old_config_id = game.game_config_id
        if old_config_id:
            new_config_id = duplicate_game_config(slugify(new_name), old_config_id)
        else:
            new_config_id = None
        categories = game.get_categories()

        duplicate_game_dialog.destroy()
        db_game = get_game_by_field(game.id, "id")
        db_game["name"] = new_name
        db_game["slug"] = slugify(new_name) if new_name != game.name else game.slug
        db_game["lastplayed"] = None
        db_game["playtime"] = 0.0
        db_game["configpath"] = new_config_id
        db_game.pop("id")
        # Disconnect duplicate from service- there should be at most 1 database game for a service game.
        db_game.pop("service", None)
        db_game.pop("service_id", None)

        game_id = add_game(**db_game)

        new_game = Game(game_id)

        # add categories before the save, so it can emit the signal. add_game()
        # means the game is already on the database, so this is legit.
        for cat in categories:
            new_game.add_category(cat, no_signal=True)

        new_game.save()

        # Download in the background; we'll update the LutrisWindow when this
        # completes, no need to wait for it.
        AsyncCall(download_lutris_media, None, db_game["slug"])

    def _select_game_launch_config_name(self, game):
        game_config = game.config.game_level.get("game", {})
        configs = game_config.get("launch_configs")

        if not configs:
            return ""  # use primary configuration

        dlg = dialogs.LaunchConfigSelectDialog(game, configs, title=_("Select shortcut target"), parent=self.window)
        if not dlg.confirmed:
            return None  # no error here- the user cancelled out

        config_index = dlg.config_index
        return configs[config_index - 1]["name"] if config_index > 0 else ""


class ServiceGameActions(GameActions):
    """This actions class supports a single service game, which has an idiosyncratic set of
    actions."""

    def __init__(self, game: Game, window: Gtk.Window, application=None):
        super().__init__(window, application)
        self.game = game

    def get_games(self):
        return [self.game]

    def get_game_actions(self):
        return [
            ("install", _("Install"), self.on_install_clicked),
            ("add", _("Locate installed game"), self.on_locate_installed_game),
            ("view", _("View on Lutris.net"), self.on_view_game),
        ]

    def get_displayed_entries(self):
        """Return a dictionary of actions that should be shown for a game"""
        return {"install": self.is_installable, "add": self.is_installable, "view": True}


def get_game_actions(games: List[Game], window: Gtk.Window, application=None) -> GameActions:
    """Creates a GameActions instance (which may be a subclass) for the list of games given. If
    it can't figure out a suitable class, it falls back to the base GameActions class, which
    provides no actions."""
    if games:
        if len(games) == 1:
            game = games[0]
            if game.is_db_stored:
                return SingleGameActions(game, window, application)

            if game.service:
                return ServiceGameActions(game, window, application)
        elif all(g.is_db_stored for g in games):
            return MultiGameActions(games, window)

    # If given no games, or the games are not of a kind we can handle,
    # the base class acts as an empty set of actions.
    return GameActions(window, application)



import json
import os
import shutil
import subprocess
from gettext import gettext as _
from pathlib import Path
from typing import Any, Dict, List, Optional

import requests

from lutris import settings
from lutris.exceptions import MissingExecutableError
from lutris.services.base import BaseService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import system
from lutris.util.log import logger
from lutris.util.strings import slugify


class FlathubBanner(ServiceMedia):
    """Standard size of a Flathub banner"""

    service = "flathub"
    size = (128, 128)
    dest_path = os.path.join(settings.CACHE_DIR, "flathub/banners")
    file_patterns = ["%s.png"]
    url_field = "icon"

    def get_media_url(self, details: Dict[str, Any]) -> Optional[str]:
        return details.get(self.url_field)


class FlathubGame(ServiceGame):
    """Representation of a Flathub game"""

    service = "flathub"

    @classmethod
    def new_from_flathub_game(cls, flathub_game):
        """Return a Flathub game instance from the API info"""
        service_game = FlathubGame()
        service_game.appid = flathub_game["app_id"]
        service_game.slug = slugify(flathub_game["name"])
        service_game.lutris_slug = slugify(flathub_game["name"])
        service_game.name = flathub_game["name"]
        service_game.runner = "flatpak"
        service_game.details = json.dumps(flathub_game)
        return service_game


class FlathubService(BaseService):
    """Service class for Flathub"""

    id = "flathub"
    name = _("Flathub")
    icon = "flathub"
    medias = {"banner": FlathubBanner}
    default_format = "banner"
    api_url = "https://flathub.org/api/v2/collection/category/game"
    cache_path = os.path.join(settings.CACHE_DIR, "flathub-library.json")

    branch = "stable"
    arch = "x86_64"
    install_type = "user"  # Should default to user. Lutris has no business installing system packages
    install_locations = {"system": "var/lib/flatpak/app/", "user": f"{Path.home()}/.local/share/flatpak/app/"}
    runner = "flatpak"
    game_class = FlathubGame

    def wipe_game_cache(self):
        """Wipe the game cache, allowing it to be reloaded"""
        if system.path_exists(self.cache_path):
            logger.debug("Deleting %s cache %s", self.id, self.cache_path)
            os.remove(self.cache_path)
        super().wipe_game_cache()

    def get_flatpak_cmd(self):
        flatpak_abspath = shutil.which("flatpak")
        if flatpak_abspath:
            return [flatpak_abspath]
        flatpak_spawn_abspath = shutil.which("flatpak-spawn")
        if flatpak_spawn_abspath:
            return [flatpak_spawn_abspath, "--host", "flatpak"]
        raise MissingExecutableError(_("No flatpak or flatpak-spawn found"))

    def load(self):
        """Load the available games from Flathub"""
        response = requests.get(self.api_url, timeout=5)
        response.raise_for_status()
        json = response.json()
        entries = json.get("hits") or []
        flathub_games = []
        for game in entries:
            flathub_games.append(FlathubGame.new_from_flathub_game(game))
        for game in flathub_games:
            game.save()
        return flathub_games

    def install(self, db_game):
        """Install a Flathub game"""
        app_id = db_game["appid"]
        logger.debug("Installing %s from service %s", app_id, self.id)
        # Check if Flathub repo is active on the system
        if not self.is_flathub_remote_active():
            raise RuntimeError(
                _("Flathub is not configured on the system. Visit https://flatpak.org/setup/ for instructions.")
            )
        # Install the game
        self.install_from_api(db_game, app_id)

    def get_installed_apps(self):
        """Get list of installed Flathub apps"""
        try:
            flatpak_cmd = self.get_flatpak_cmd()
            process = subprocess.run(
                flatpak_cmd + ["list", "--app", "--columns=application"],
                capture_output=True,
                check=True,
                encoding="utf-8",
                text=True,
                timeout=5.0,
            )
            return process.stdout.splitlines() or []
        except (TimeoutError, subprocess.CalledProcessError) as err:
            logger.exception("Error occurred while getting installed flatpak apps: %s", err)
            return []

    def is_flathub_remote_active(self):
        """Check if Flathub is configured and enabled as a flatpak repository"""
        remotes = self.get_active_remotes()
        if not remotes:
            logger.warning("Remotes not found, Flathub considered installed")
            return True
        for remote in remotes:
            if "flathub" in remote.values():
                return True
        return False

    def get_active_remotes(self):
        """Get a list of dictionaries containing name, title and url"""
        try:
            flatpak_cmd = self.get_flatpak_cmd()
            process = subprocess.run(
                flatpak_cmd + ["remotes", "--columns=name,title,url"],
                capture_output=True,
                check=True,
                encoding="utf-8",
                text=True,
                timeout=5.0,
            )
            entries = []
            for line in process.stdout.splitlines():
                cols = line.split("\t")
                entries.append({"name": cols[0].lower(), "title": cols[1].lower(), "url": cols[2]})
            return entries
        except (TimeoutError, subprocess.CalledProcessError) as err:
            logger.exception("Error occurred while getting installed flatpak apps: %s", err)
            return []

    def generate_installer(self, db_game):
        # TODO: Add options for user to select arch, branch and install_type
        flatpak_cmd = self.get_flatpak_cmd()
        return {
            "appid": db_game["appid"],
            "game_slug": self.get_installed_slug(db_game),
            "slug": slugify(db_game["name"]) + "-" + self.id,
            "name": db_game["name"],
            "version": "Flathub",
            "runner": self.get_installed_runner_name(db_game),
            "script": {
                "game": {
                    "appid": db_game["appid"],
                    "arch": self.arch,
                    "branch": self.branch,
                    "install_type": self.install_type,
                },
                "system": {"disable_runtime": True},
                "require-binaries": flatpak_cmd[0],
                "installer": [
                    {
                        "execute": {
                            "file": flatpak_cmd[0],
                            "args": " ".join(flatpak_cmd[1:])
                            + f" install --{self.install_type} --app --noninteractive flathub "
                            f"app/{db_game['appid']}/{self.arch}/{self.branch}",
                            "disable_runtime": True,
                        }
                    }
                ],
            },
        }

    def get_installed_runner_name(self, db_game):
        return self.runner

    def get_game_directory(self, _installer):
        install_type, application, arch, branch = (
            _installer["script"]["game"][key] for key in ("install_type", "application", "arch", "branch")
        )
        return os.path.join(self.install_locations[install_type], application, arch, branch)

    def get_game_platforms(self, db_game: dict) -> List[str]:
        return ["Linux"]

    # def add_installed_games(self):
    #     process = subprocess.run(["flatpak", "list", "--app", "--columns=application,arch,branch,installation,name"],
    #                              capture_output=True, check=True, encoding="utf-8", text=True)
    #     for line in process.stdout.splitlines():
    #         cols = line.split("\t")



import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from lutris import settings
from lutris.database import sql
from lutris.gui.widgets import NotificationSource

SAVED_SEARCHES_UPDATED = NotificationSource()


@dataclass
class SavedSearch:
    saved_search_id: int
    name: str
    search: str

    def add(self, no_signal: bool = False) -> None:
        """Add a category to the database"""
        self.saved_search_id = sql.db_insert(
            settings.DB_PATH, "saved_searches", {"name": self.name, "search": self.search}
        )
        if not no_signal:
            SAVED_SEARCHES_UPDATED.fire()

    def update(self, no_signal: bool = False) -> None:
        query = "UPDATE saved_searches SET name=?, search=? WHERE id=?"

        with sql.db_cursor(settings.DB_PATH) as cursor:
            sql.cursor_execute(cursor, query, (self.name, self.search, self.saved_search_id))
        if not no_signal:
            SAVED_SEARCHES_UPDATED.fire()

    def remove(self, no_signal: bool = False) -> None:
        query = "DELETE FROM saved_searches WHERE id=?"

        with sql.db_cursor(settings.DB_PATH) as cursor:
            sql.cursor_execute(cursor, query, (self.saved_search_id,))

        if not no_signal:
            SAVED_SEARCHES_UPDATED.fire()


def _create_search(row: Dict[str, Any]) -> "SavedSearch":
    return SavedSearch(row["id"], row["name"], row["search"])


def strip_saved_search_name(name):
    """This strips the name given, and also removes extra internal whitespace."""
    name = (name or "").strip()
    name = re.sub(" +", " ", name)  # Remove excessive whitespaces
    return name


def get_saved_searches() -> List[SavedSearch]:
    """Get the list of every search in database."""
    rows = sql.db_select(settings.DB_PATH, "saved_searches")
    return [_create_search(row) for row in rows]


def get_saved_search_by_name(name: str) -> Optional[SavedSearch]:
    """Return a category by name"""
    categories = sql.db_select(settings.DB_PATH, "saved_searches", condition=("name", name))
    if categories:
        return _create_search(categories[0])

    return None


def get_saved_search_by_id(saved_search_id: int) -> Optional[SavedSearch]:
    """Return a category by name"""
    categories = sql.db_select(settings.DB_PATH, "saved_searches", condition=("id", saved_search_id))
    if categories:
        return _create_search(categories[0])

    return None



"""Module for handling the PGA cache"""

import os
import shutil
from gettext import gettext as _
from typing import Optional, Tuple
from urllib.parse import urlparse

from lutris import settings
from lutris.util.log import logger
from lutris.util.system import merge_folders, path_contains


def get_cache_path(create: bool = False) -> str:
    """Returns the directory under which Lutris caches install files. This can be specified
    by the user, but defaults to a location in ~/.cache."""
    cache_path = get_custom_cache_path()
    if cache_path:
        cache_path = os.path.expanduser(cache_path)
        if os.path.isdir(cache_path) or os.path.isdir(os.path.dirname(cache_path)):
            return cache_path

    return settings.INSTALLER_CACHE_DIR


def get_url_cache_path(url: str, file_id: str, game_slug: str, prepare: bool = False) -> str:
    """Return the directory used as a cache a file that will be downloaded from
    a URL. You also provide the file id from the file list and the slug for
    the game.

    The files will be cached in a directory named after the slug, but
    in a further subdirectory for the file-id; since GOG file-ids are
    inconvenient for this, we special case URLs pointed to 'gog.com' here.

    If 'prepare' is true, this will also create the directory."""
    cache_path = get_cache_path()
    url_parts = urlparse(url)
    if url_parts.netloc.endswith("gog.com"):
        folder = "gog"
    else:
        folder = file_id
    path = os.path.join(cache_path, game_slug, folder)

    if prepare:
        if not os.path.exists(path):
            os.makedirs(path)

    return path


def get_custom_cache_path() -> Optional[str]:
    """Returns the custom path, wether it is usable or not. Returns
    None if the path is not set, so that the default INSTALLER_CACHE_DIR
    should be used."""
    cache_path = settings.read_setting("pga_cache_path")
    return cache_path if cache_path else None


def has_valid_custom_cache_path() -> bool:
    """True if the custom cache path is set and refers to a usable
    directory, so that get_cache_path() will return it. The directory
    does not have to exist for this, but at least its *parent* directory
    does."""
    cache_path = get_custom_cache_path()
    if not cache_path:
        return False

    valid, msg = validate_custom_cache_path(cache_path)
    if msg:
        logger.warning(msg)

    return valid


def validate_custom_cache_path(cache_path: str) -> Tuple[bool, Optional[str]]:
    """Checks the validity of a given path; returns a flag for whether
    it can be used, and an optional message for what's wrong with it,
    if anything is."""
    cache_path = os.path.expanduser(cache_path)
    if not os.path.isdir(cache_path):
        parent = os.path.dirname(cache_path)
        if os.path.isdir(parent):
            return True, _(
                "The cache path '%s' does not exist, but its parent does so it will be created when needed."
            ) % cache_path
        else:
            return False, _(
                "The cache path %s does not exist, nor does its parent, so it won't be created."
            ) % cache_path

    return True, None


def save_custom_cache_path(path: str) -> None:
    """Saves the PGA cache path to the settings"""
    settings.write_setting("pga_cache_path", path)


def is_file_in_custom_cache(path: str) -> bool:
    """True if the 'path' is inside the custom cache (so we should
    not causally delete it). False for files in INSTALLER_CACHE_DIR -
    that is a cache, but not the custom cache, and we do delete those
    files freely."""
    cache_path = get_custom_cache_path()
    return bool(cache_path and path_contains(cache_path, path))


def save_to_cache(source: str, destination: str) -> None:
    """Copy a file or folder to the cache"""
    if not source:
        raise ValueError("Missing source")
    if os.path.dirname(source) == destination:
        logger.info("File %s is already cached in %s", source, destination)
        return
    if os.path.isdir(source):
        # Copy folder recursively
        merge_folders(source, destination)
    else:
        shutil.copy(source, destination)
    logger.debug("Cached %s to %s", source, destination)



"""Widgets for the installer window"""

import os
from gettext import gettext as _

from gi.repository import GObject, Gtk

from lutris.gui.installer.widgets import InstallerLabel
from lutris.gui.widgets.common import FileChooserEntry
from lutris.installer.installer_file_collection import InstallerFileCollection
from lutris.installer.steam_installer import SteamInstaller
from lutris.util import system
from lutris.util.log import logger
from lutris.util.strings import gtk_safe


class InstallerFileBox(Gtk.VBox):
    """Container for an installer file downloader / selector"""

    __gsignals__ = {
        "file-available": (GObject.SIGNAL_RUN_FIRST, None, ()),
        "file-ready": (GObject.SIGNAL_RUN_FIRST, None, ()),
        "file-unready": (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self, installer_file):
        super().__init__()
        self.installer_file = installer_file
        self.cache_to_pga = self.installer_file.uses_pga_cache()
        self.started = False
        self.start_func = None
        self.stop_func = None
        self.state_label = None  # Use this label to display status update
        self.set_margin_left(12)
        self.set_margin_right(12)
        self.provider = self.installer_file.default_provider
        self.file_provider_widget = None
        self.add(self.get_widgets())

    @property
    def is_ready(self):
        """Whether the file is ready to be downloaded / fetched from its provider"""
        return self.installer_file.is_ready(self.provider)

    def get_download_progress(self):
        """Return the widget for the download progress bar"""
        download_progress = self.installer_file.create_download_progress_box()
        download_progress.connect("complete", self.on_download_complete)
        download_progress.connect("cancel", self.on_download_cancelled)
        download_progress.connect("error", self.on_download_error)
        download_progress.show()
        self.installer_file.remove_previous()
        return download_progress

    def get_file_provider_widget(self):
        """Return the widget used to track progress of file"""
        box = Gtk.VBox(spacing=6)
        if self.provider == "download":
            download_progress = self.get_download_progress()
            self.start_func = download_progress.start
            self.stop_func = download_progress.on_cancel_clicked
            box.pack_start(download_progress, False, False, 0)
            return box
        if self.provider == "pga":
            url_label = InstallerLabel("In cache: %s" % self.installer_file.get_label(), wrap=False)
            box.pack_start(url_label, False, False, 6)
            return box
        if self.provider == "user":
            user_label = InstallerLabel(gtk_safe(self.installer_file.human_url))
            box.pack_start(user_label, False, False, 0)
            return box
        # InstallerFileCollection should not have steam provider
        if self.provider == "steam":
            if isinstance(self.installer_file, InstallerFileCollection):
                raise RuntimeError("Installer file is type InstallerFileCollection and do not support 'steam' provider")
            steam_installer = SteamInstaller(self.installer_file.url, self.installer_file.id)
            steam_installer.game_installed.register(self.on_steam_game_installed)
            steam_installer.game_state_changed.register(self.on_steam_game_state_changed)
            self.start_func = steam_installer.install_steam_game
            self.stop_func = steam_installer.stop_func

            steam_box = Gtk.HBox(spacing=6)
            info_box = Gtk.VBox(spacing=6)
            steam_label = InstallerLabel(_("Steam game <b>{appid}</b>").format(appid=steam_installer.appid))
            info_box.add(steam_label)
            self.state_label = InstallerLabel("")
            info_box.add(self.state_label)
            steam_box.add(info_box)
            return steam_box
        raise ValueError("Invalid provider %s" % self.provider)

    def get_combobox_model(self):
        """ "Return the combobox's model"""
        model = Gtk.ListStore(str, str)
        if "download" in self.installer_file.providers:
            model.append(["download", _("Download")])
        if "pga" in self.installer_file.providers:
            model.append(["pga", _("Use Cache")])
        if "steam" in self.installer_file.providers:
            model.append(["steam", _("Steam")])
        if not (isinstance(self.installer_file, InstallerFileCollection) and self.installer_file.service == "amazon"):
            model.append(["user", _("Select File")])
        return model

    def get_combobox(self):
        """Return the combobox widget to select file source"""
        combobox = Gtk.ComboBox.new_with_model(self.get_combobox_model())
        combobox.set_id_column(0)
        renderer_text = Gtk.CellRendererText()
        combobox.pack_start(renderer_text, True)
        combobox.add_attribute(renderer_text, "text", 1)
        combobox.connect("changed", self.on_source_changed)
        combobox.set_active_id(self.provider)
        return combobox

    def replace_file_provider_widget(self):
        """Replace the file provider label and the source button with the actual widget"""
        self.file_provider_widget.destroy()
        widget_box = self.get_children()[0]
        if self.started:
            self.file_provider_widget = self.get_file_provider_widget()
            # Also remove the the source button
            for child in widget_box.get_children():
                child.destroy()
        else:
            self.file_provider_widget = self.get_file_provider_label()
        widget_box.pack_start(self.file_provider_widget, True, True, 0)
        widget_box.reorder_child(self.file_provider_widget, 0)
        widget_box.show_all()

    def on_source_changed(self, combobox):
        """Change the source to a new provider, emit a new state"""
        tree_iter = combobox.get_active_iter()
        if tree_iter is None:
            return
        model = combobox.get_model()
        source = model[tree_iter][0]
        if source == self.provider:
            return
        self.provider = source
        self.replace_file_provider_widget()
        if self.provider == "user":
            self.emit("file-unready")
        else:
            self.emit("file-ready")

    def get_file_provider_label(self):
        """Return the label displayed before the download starts"""
        if self.provider == "user":
            box = Gtk.VBox(spacing=6)
            label = InstallerLabel(self.installer_file.get_label())
            label.props.can_focus = True
            box.pack_start(label, False, False, 0)
            location_entry = FileChooserEntry(self.installer_file.human_url, Gtk.FileChooserAction.OPEN)
            location_entry.connect("changed", self.on_location_changed)
            location_entry.show()
            box.pack_start(location_entry, False, False, 0)
            if self.installer_file.is_user_pga_caching_allowed:
                cache_option = Gtk.CheckButton(_("Cache file for future installations"))
                cache_option.set_active(self.cache_to_pga)
                cache_option.connect("toggled", self.on_user_file_cached)
                box.pack_start(cache_option, False, False, 0)
            return box
        return InstallerLabel(self.installer_file.get_label())

    def get_widgets(self):
        """Return the widget with the source of the file and a way to change its source"""
        box = Gtk.HBox(spacing=12, margin_top=6, margin_bottom=6)
        self.file_provider_widget = self.get_file_provider_label()
        box.pack_start(self.file_provider_widget, True, True, 0)
        source_box = Gtk.HBox()
        source_box.props.valign = Gtk.Align.START
        box.pack_start(source_box, False, False, 0)

        aux_info = self.installer_file.auxiliary_info
        if aux_info:
            source_box.pack_start(InstallerLabel(aux_info), False, False, 0)
        source_box.pack_start(InstallerLabel(_("Source:")), False, False, 0)
        combobox = self.get_combobox()
        source_box.pack_start(combobox, False, False, 0)
        return box

    def on_location_changed(self, widget):
        """Open a file picker when the browse button is clicked"""
        file_path = os.path.expanduser(widget.get_text())
        self.installer_file.override_dest_file(file_path)
        if system.path_exists(file_path):
            self.emit("file-ready")
        else:
            self.emit("file-unready")

    def on_user_file_cached(self, checkbutton):
        """Enable or disable caching of user provided files"""
        self.cache_to_pga = checkbutton.get_active()

    def on_state_changed(self, _widget, state):
        """Update the state label with a new state"""
        self.state_label.set_text(state)

    def on_steam_game_state_changed(self, installer):
        """Update the state label with a new state"""
        self.state_label.set_text(installer.state)

    def start(self):
        """Starts the download of the file"""
        self.started = True
        self.installer_file.prepare()
        self.replace_file_provider_widget()
        if self.provider in ("pga", "user") and self.is_ready:
            self.emit("file-available")
            self.cache_file()
            return
        if self.start_func:
            return self.start_func()

    def cache_file(self):
        """Copy file to the PGA cache"""
        if self.cache_to_pga:
            self.installer_file.save_to_cache()

    def on_download_cancelled(self, downloader):
        """Handle cancellation of installers"""
        logger.error("Download from %s cancelled", downloader)
        downloader.set_retry_button()

    def on_download_error(self, downloader, error):
        logger.error("Download from %s failed: %s", downloader, error)
        downloader.set_retry_button()

    def on_download_complete(self, widget, _data=None):
        """Action called on a completed download."""
        logger.info("Download completed")
        self.installer_file.check_hash()
        if isinstance(widget, SteamInstaller):
            self.installer_file.dest_file = widget.get_steam_data_path()
        else:
            self.cache_file()
        self.emit("file-available")

    def on_steam_game_installed(self, installer):
        self.installer_file.dest_file = installer.get_steam_data_path()
        self.emit("file-available")



import os

from lutris.api import get_api_games, get_game_installers
from lutris.database.games import get_games
from lutris.installer.errors import MissingGameDependencyError
from lutris.installer.interpreter import ScriptInterpreter
from lutris.services.lutris import download_lutris_media
from lutris.util.log import logger
from lutris.util.strings import slugify


def get_game_slugs_and_folders(dirname):
    """Scan a directory for games previously installed with lutris"""
    folders = os.listdir(dirname)
    game_folders = {}
    for folder in folders:
        if not os.path.isdir(os.path.join(dirname, folder)):
            continue
        game_folders[slugify(folder)] = folder
    return game_folders


def find_game_folder(dirname, api_game, slugs_map):
    if api_game["slug"] in slugs_map:
        game_folder = os.path.join(dirname, slugs_map[api_game["slug"]])
        if os.path.exists(game_folder):
            return game_folder
    for alias in api_game["aliases"]:
        if alias["slug"] in slugs_map:
            game_folder = os.path.join(dirname, slugs_map[alias["slug"]])
            if os.path.exists(game_folder):
                return game_folder


def detect_game_from_installer(game_folder, installer):
    try:
        exe_path = installer["script"]["game"].get("exe")
    except KeyError:
        exe_path = installer["script"].get("exe")
    if not exe_path:
        try:
            exe_path = installer["script"]["game"].get("main_file")
        except KeyError:
            pass
    if not exe_path:
        return None
    exe_path = exe_path.replace("$GAMEDIR/", "")
    full_path = os.path.join(game_folder, exe_path)
    if os.path.exists(full_path):
        return full_path


def find_game(game_folder, api_game):
    installers = get_game_installers(api_game["slug"])
    for installer in installers:
        full_path = detect_game_from_installer(game_folder, installer)
        if full_path:
            return full_path, installer
    return None, None


def get_used_directories():
    directories = set()
    for game in get_games():
        if game["directory"]:
            directories.add(game["directory"])
    return directories


def install_game(installer, game_folder):
    interpreter = ScriptInterpreter(installer)
    interpreter.target_path = game_folder
    interpreter.installer.save()


def scan_directory(dirname):
    slugs_map = get_game_slugs_and_folders(dirname)
    directories = get_used_directories()
    api_games = get_api_games(list(slugs_map.keys()))
    slugs_seen = set()
    slugs_installed = set()
    for api_game in api_games:
        if api_game["slug"] in slugs_seen:
            continue
        slugs_seen.add(api_game["slug"])
        game_folder = find_game_folder(dirname, api_game, slugs_map)
        if game_folder in directories:
            slugs_installed.add(api_game["slug"])
            continue
        full_path, installer = find_game(game_folder, api_game)
        if full_path:
            logger.info("Found %s in %s", api_game["name"], full_path)
            try:
                install_game(installer, game_folder)
            except MissingGameDependencyError as ex:
                logger.error("Skipped %s: %s", api_game["name"], ex)
            download_lutris_media(installer["game_slug"])
            slugs_installed.add(api_game["slug"])

    installed_map = {slug: folder for slug, folder in slugs_map.items() if slug in slugs_installed}
    missing_map = {slug: folder for slug, folder in slugs_map.items() if slug not in slugs_installed}
    return installed_map, missing_map



import os

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")

from lutris import startup
from lutris.database import schema


def setup_test_environment():
    """Sets up a system to be able to run tests"""
    os.environ["LUTRIS_SKIP_INIT"] = "1"
    schema.syncdb()
    startup.init_lutris()



"""Window to show game logs"""

import os
from datetime import datetime
from gettext import gettext as _

from gi.repository import Gdk, GObject, Gtk, Pango

from lutris.gui.dialogs import FileDialog
from lutris.gui.widgets.log_text_view import LogTextView
from lutris.util import datapath


class LogWindow(GObject.Object):
    def __init__(self, game, buffer, application=None):
        super().__init__()
        ui_filename = os.path.join(datapath.get(), "ui/log-window.ui")
        builder = Gtk.Builder()
        builder.add_from_file(ui_filename)
        builder.connect_signals(self)
        self.window = builder.get_object("log_window")

        self.title = _("Log for {}").format(game)
        self.window.set_title(self.title)

        self.buffer = buffer
        self.logtextview = LogTextView(self.buffer)

        scrolled_window = builder.get_object("scrolled_window")
        scrolled_window.add(self.logtextview)

        self.search_entry = builder.get_object("search_entry")
        self.search_entry.connect("search-changed", self.logtextview.find_first)
        self.search_entry.connect("next-match", self.logtextview.find_next)
        self.search_entry.connect("previous-match", self.logtextview.find_previous)

        save_button = builder.get_object("save_button")
        save_button.connect("clicked", self.on_save_clicked)

        # Add zoom buttons
        zoom_in_button = builder.get_object("zoom_in_button")
        zoom_out_button = builder.get_object("zoom_out_button")
        zoom_in_button.connect("clicked", self.on_zoom_in_clicked)
        zoom_out_button.connect("clicked", self.on_zoom_out_clicked)

        self.window.connect("key-press-event", self.on_key_press_event)
        self.window.show_all()

    def on_key_press_event(self, widget, event):
        shift = event.state & Gdk.ModifierType.SHIFT_MASK
        if event.keyval == Gdk.KEY_Return:
            if shift:
                self.search_entry.emit("previous-match")
            else:
                self.search_entry.emit("next-match")

    def on_save_clicked(self, _button):
        """Handler to save log to a file"""
        now = datetime.now()
        log_filename = "%s (%s).log" % (self.title, now.strftime("%Y-%m-%d-%H-%M"))
        file_dialog = FileDialog(
            message="Save the logs to...", default_path=os.path.expanduser("~/%s" % log_filename), mode="save"
        )
        log_path = file_dialog.filename
        if not log_path:
            return

        text = self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True)
        with open(log_path, "w", encoding="utf-8") as log_file:
            log_file.write(text)

    def on_zoom_in_clicked(self, _button):
        """Increase font size"""
        font = self.logtextview.get_style_context().get_font(Gtk.StateFlags.NORMAL)
        size = font.get_size() / Pango.SCALE
        if size < 48:  # Maximum size
            self.logtextview.override_font(Pango.FontDescription(f"monospace {size + 1}"))

    def on_zoom_out_clicked(self, _button):
        """Decrease font size"""
        font = self.logtextview.get_style_context().get_font(Gtk.StateFlags.NORMAL)
        size = font.get_size() / Pango.SCALE
        if size > 6:  # Minimum size
            self.logtextview.override_font(Pango.FontDescription(f"monospace {size - 1}"))



"""Check to run at program start"""

import os
import sqlite3
from gettext import gettext as _

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")

from gi.repository import GdkPixbuf

from lutris import runners, settings
from lutris.database.games import get_games
from lutris.database.schema import syncdb
from lutris.game import Game
from lutris.runners.json import load_json_runners
from lutris.services import DEFAULT_SERVICES
from lutris.util.graphics import vkquery
from lutris.util.graphics.drivers import get_gpu_cards
from lutris.util.graphics.gpu import GPU, GPUS
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger
from lutris.util.path_cache import build_path_cache
from lutris.util.system import create_folder
from lutris.util.wine.dxvk import REQUIRED_VULKAN_API_VERSION


def init_dirs():
    """Creates Lutris directories"""
    directories = [
        settings.CONFIG_DIR,
        settings.RUNNERS_CONFIG_DIR,
        settings.GAME_CONFIG_DIR,
        settings.DATA_DIR,
        settings.ICON_PATH,
        settings.BANNER_PATH,
        settings.COVERART_PATH,
        settings.RUNNER_DIR,
        settings.RUNTIME_DIR,
        settings.CACHE_DIR,
        settings.SHADER_CACHE_DIR,
        settings.INSTALLER_CACHE_DIR,
        settings.TMP_DIR,
    ]
    for directory in directories:
        create_folder(directory)


def check_libs(all_components=False):
    """Checks that required libraries are installed on the system"""
    missing_libs = LINUX_SYSTEM.get_missing_libs()
    if all_components:
        components = LINUX_SYSTEM.requirements
    else:
        components = LINUX_SYSTEM.critical_requirements

    for req in components:
        for index, arch in enumerate(LINUX_SYSTEM.runtime_architectures):
            for lib in missing_libs[req][index]:
                logger.error("%s %s missing (needed by %s)", arch, lib, req.lower())


def check_vulkan():
    """Reports if Vulkan is enabled on the system"""
    if os.environ.get("LUTRIS_NO_VKQUERY"):
        return
    if not vkquery.is_vulkan_supported():
        logger.warning("Vulkan is not available or your system isn't Vulkan capable")
    else:
        required_api_version = REQUIRED_VULKAN_API_VERSION
        library_api_version = vkquery.get_vulkan_api_version()
        if library_api_version and library_api_version < required_api_version:
            logger.warning(
                "Vulkan reports an API version of %s. %s is required for the latest DXVK.",
                vkquery.format_version(library_api_version),
                vkquery.format_version(required_api_version),
            )

        devices = vkquery.get_device_info()

        if devices and devices[0].api_version < required_api_version:
            logger.warning(
                "Vulkan reports that the '%s' device has API version of %s. %s is required for the latest DXVK.",
                devices[0].name,
                vkquery.format_version(devices[0].api_version),
                vkquery.format_version(required_api_version),
            )


def check_gnome():
    required_names = ["svg", "png", "jpeg"]
    format_names = [f.get_name() for f in GdkPixbuf.Pixbuf.get_formats()]
    for required in required_names:
        if required not in format_names:
            logger.error("'%s' PixBuf support is not installed.", required.upper())


def fill_missing_platforms():
    """Sets the platform on games where it's missing.
    This should never happen.
    """
    pga_games = get_games(filters={"installed": 1})
    for pga_game in pga_games:
        if pga_game.get("platform") or not pga_game["runner"]:
            continue
        game = Game(game_id=pga_game["id"])
        game.set_platform_from_runner()
        if game.platform:
            logger.info("Platform for %s set to %s", game.name, game.platform)
            game.save_platform()


def run_all_checks() -> None:
    """Run all startup checks"""
    for card in get_gpu_cards():
        gpu = GPU(card)
        driver_info = gpu.get_driver_info()
        logger.info('"%s" is %s Driver %s', card, gpu, driver_info.get("version"))
        GPUS[card] = gpu

    check_libs()
    check_vulkan()
    check_gnome()
    fill_missing_platforms()
    build_path_cache()


def init_lutris():
    """Run full initialization of Lutris"""
    runners.inject_runners(load_json_runners())
    init_dirs()
    try:
        syncdb()
    except sqlite3.DatabaseError as err:
        raise RuntimeError(
            _("Failed to open database file in %s. Try renaming this file and relaunch Lutris") % settings.DB_PATH
        ) from err
    for service in DEFAULT_SERVICES:
        if not settings.read_setting(service, section="services"):
            settings.write_setting(service, True, section="services")



"""Install a game by following its install script."""

import os
from gettext import gettext as _

from gi.repository import GObject

from lutris import settings
from lutris.config import LutrisConfig
from lutris.database.games import get_game_by_field
from lutris.exceptions import AuthenticationError, MisconfigurationError, UnavailableGameError
from lutris.gui.dialogs.delegates import Delegate
from lutris.installer import AUTO_EXE_PREFIX
from lutris.installer.commands import CommandsMixin
from lutris.installer.errors import MissingGameDependencyError, ScriptingError
from lutris.installer.installer import LutrisInstaller
from lutris.runners import NonInstallableRunnerError, RunnerInstallationError, steam, wine
from lutris.services.lutris import download_lutris_media
from lutris.util import system
from lutris.util.display import DISPLAY_MANAGER
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger
from lutris.util.strings import unpack_dependencies


class ScriptInterpreter(GObject.Object, CommandsMixin):
    """Control the execution of an installer"""

    __gsignals__ = {
        "runners-installed": (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    class InterpreterUIDelegate(Delegate):
        """This is a base class for objects that provide UI services
        for running scripts. The InstallerWindow inherits from this."""

        def __init__(self, service=None, appid=None):
            self.service = service
            self.appid = appid

        def report_error(self, error):
            """Called to report an error during installation. The installation will then stop."""
            pass

        def report_status(self, status):
            """Called to report the current activity of the installer."""

        def attach_log(self, command):
            """Called to attach the command to a log UI, so its log output can be viewed."""

        def begin_disc_prompt(self, message, requires, installer, callback):
            """Called to prompt for a disc. When the disc is provided, the callback is invoked.
            The method returns immediately, however."""
            raise NotImplementedError()

        def begin_input_menu(self, alias, options, preselect, callback):
            """Called to prompt the user to select among a list of options. When the user
            does so, the callback is invoked. The method returns immediately, however."""
            raise NotImplementedError()

        def report_finished(self, game_id, status):
            """Called to report the successful completion of the installation."""
            logger.info("Installation of game %s completed.", game_id)

    def __init__(self, installer, interpreter_ui_delegate=None):
        super().__init__()
        self.target_path = None
        self.interpreter_ui_delegate = interpreter_ui_delegate or ScriptInterpreter.InterpreterUIDelegate()
        self.service = self.interpreter_ui_delegate.service
        _appid = self.interpreter_ui_delegate.appid
        self.game_dir_created = False  # Whether a game folder was created during the install
        # Extra files for installers, either None if the extras haven't been checked yet.
        # Or a list of IDs of extras to be downloaded during the install
        self.game_disc = None
        self.game_files = {}
        self.cancelled = False
        self.abort_current_task = None
        self.user_inputs = []
        self.current_command = 0  # Current installer command when iterating through them
        self.runners_to_install = []
        self.current_resolution = DISPLAY_MANAGER.get_current_resolution()
        self.installer = LutrisInstaller(installer, self, service=self.service, appid=_appid)

        if not self.installer.script:
            raise ScriptingError(_("This installer doesn't have a 'script' section"))
        if not self.service and self.installer.service:
            self.service = self.installer.service
        script_errors = self.installer.get_errors()
        if script_errors:
            raise ScriptingError(
                _("Invalid script: \n{}").format("\n".join(script_errors)), faulty_data=self.installer.script
            )

        self._check_binary_dependencies()
        self._check_dependency()
        if self.installer.creates_game_folder:
            self.target_path = self.get_default_target()

    def on_timeout_error(self, error):
        self.interpreter_ui_delegate.report_error(error)

    def on_idle_error(self, error):
        self.interpreter_ui_delegate.report_error(error)

    def on_signal_error(self, error):
        self.interpreter_ui_delegate.report_error(error)

    def on_emission_hook_error(self, error):
        self.interpreter_ui_delegate.report_error(error)

    @property
    def appid(self):
        logger.warning("Do not access appid from interpreter")
        return self.installer.service_appid

    def get_default_target(self):
        """Return default installation dir"""
        config = LutrisConfig(runner_slug=self.installer.runner)
        games_dir = config.system_config.get("game_path", os.path.expanduser("~"))
        if self.service:
            service_dir = self.service.id
        else:
            service_dir = ""
        return os.path.expanduser(os.path.join(games_dir, service_dir, self.installer.game_slug))

    @property
    def cache_path(self):
        """Return the directory used as a cache for the duration of the installation"""
        return os.path.join(settings.INSTALLER_CACHE_DIR, "%s" % self.installer.game_slug)

    @property
    def script_env(self):
        """Return the script's own environment variable with values
        susbtituted. This value can be used to provide the same environment
        variable as set for the game during the install process.
        """
        return {
            key: self._substitute(value)
            for key, value in self.installer.script.get("system", {}).get("env", {}).items()
        }

    @staticmethod
    def _get_game_dependency(dependency):
        """Return a game database row from a dependency name"""
        game = get_game_by_field(dependency, field="installer_slug")
        if not game:
            game = get_game_by_field(dependency, "slug")

        # Game must be installed and have a directory
        # set so we can use that as the destination
        if game and game["installed"] and game["directory"]:
            return game

    def _check_binary_dependencies(self):
        """Check if all required binaries are installed on the system.

        This reads a `require-binaries` entry in the script, parsed the same way as
        the `requires` entry.
        """
        binary_dependencies = unpack_dependencies(self.installer.script.get("require-binaries"))
        for dependency in binary_dependencies:
            if isinstance(dependency, tuple):
                installed_binaries = {
                    dependency_option: system.can_find_executable(dependency_option) for dependency_option in dependency
                }
                if not any(installed_binaries.values()):
                    raise ScriptingError(_("This installer requires %s on your system") % _(" or ").join(dependency))
            else:
                if not system.can_find_executable(dependency):
                    raise ScriptingError(_("This installer requires %s on your system") % dependency)

    def _check_dependency(self):
        """When a game is a mod or an extension of another game, check that the base
        game is installed.
        If the game is available, install the game in the base game folder.
        The first game available listed in the dependencies is the one picked to base
        the installed on.
        """
        if self.installer.extends:
            dependencies = [self.installer.extends]
        else:
            dependencies = unpack_dependencies(self.installer.requires)
        error_message = _("You need to install {} before")
        for index, dependency in enumerate(dependencies):
            if isinstance(dependency, tuple):
                installed_games = [dep for dep in [self._get_game_dependency(dep) for dep in dependency] if dep]
                if not installed_games:
                    if len(dependency) == 1:
                        raise MissingGameDependencyError(slug=dependency)
                    raise ScriptingError(error_message.format(_(" or ").join(dependency)))
                if index == 0:
                    self.target_path = installed_games[0]["directory"]
                    self.requires = installed_games[0]["installer_slug"]
            else:
                game = self._get_game_dependency(dependency)
                if not game:
                    raise MissingGameDependencyError(slug=dependency)
                if index == 0:
                    self.target_path = game["directory"]
                    self.requires = game["installer_slug"]

    def get_extras(self):
        """Get extras and store them to move them at the end of the install"""
        if not self.service or not self.service.has_extras or not self.installer.service_appid:
            return []
        try:
            return self.service.get_extras(self.installer.service_appid)
        except (AuthenticationError, UnavailableGameError) as ex:
            logger.exception("Unable to download list of extras: %s", ex)
            return []

    def launch_install(self, ui_delegate):
        """Launch the install process; returns False if cancelled by the user."""
        self.runners_to_install = self.get_runners_to_install()
        self.install_runners(ui_delegate)
        return True

    def create_game_folder(self):
        """Create the game folder if needed and store if is was created"""
        if (
            self.installer.files
            and self.target_path
            and not system.path_exists(self.target_path)
            and self.installer.creates_game_folder
        ):
            try:
                logger.debug("Creating destination path %s", self.target_path)
                os.makedirs(self.target_path)
                self.game_dir_created = True
            except PermissionError as err:
                raise ScriptingError(
                    _("Lutris does not have the necessary permissions to install to path:"),
                    faulty_data=self.target_path,
                ) from err
            except FileNotFoundError as err:
                raise ScriptingError(
                    _("Path %s not found, unable to create game folder. Is the disk mounted?"),
                    faulty_data=self.target_path,
                ) from err

    def get_runners_to_install(self):
        """Check if the runner is installed before starting the installation
        Install the required runner(s) if necessary. This should handle runner
        dependencies or runners used for installer tasks.
        """
        runners_to_install = []
        required_runners = []
        runner = self.get_runner_class(self.installer.runner)
        required_runners.append(runner())

        for command in self.installer.script.get("installer", []):
            command_name, command_params = self._get_command_name_and_params(command)
            if command_name == "task":
                runner_name, _task_name = self._get_task_runner_and_name(command_params["name"])
                runner_names = [r.name for r in required_runners]
                if runner_name not in runner_names:
                    required_runners.append(self.get_runner_class(runner_name)())

        for runner in required_runners:
            if not runner.is_installed_for(self):
                logger.info("Runner %s needs to be installed", runner.name)
                runners_to_install.append(runner)

        return runners_to_install

    def install_runners(self, ui_delegate):
        """Install required runners for a game"""
        if self.runners_to_install:
            self.install_runner(self.runners_to_install.pop(0), ui_delegate)
            return  # install_runner calls back into this method to get the next one

        self.emit("runners-installed")

    def install_runner(self, runner, ui_delegate):
        """Install runner required by the install script"""

        def install_more_runners():
            self.install_runners(ui_delegate)

        logger.debug("Installing %s", runner.name)
        try:
            runner.install(
                ui_delegate,
                version=runner.get_installer_runner_version(self.installer),
                callback=install_more_runners,
            )
        except (NonInstallableRunnerError, RunnerInstallationError) as ex:
            logger.error(ex.message)
            raise ScriptingError.wrap(ex) from ex

    def launch_installer_commands(self):
        """Run the pre-installation steps and launch install."""
        self.create_game_folder()

        os.makedirs(self.cache_path, exist_ok=True)

        self._iter_commands()

    def _iter_commands(self, result=None, exception=None):
        if result == "STOP" or self.cancelled:
            return

        try:
            commands = self.installer.script.get("installer", [])
            if exception:
                logger.error("Last install command failed, show error")
                self.interpreter_ui_delegate.report_error(exception)
            elif self.current_command < len(commands):
                try:
                    command = commands[self.current_command]
                except KeyError as err:
                    raise ScriptingError(_("Installer commands are not formatted correctly")) from err
                self.current_command += 1
                method, params = self._map_command(command)
                if isinstance(params, dict):
                    status_text = params.pop("description", None)
                else:
                    status_text = None
                if status_text:
                    self.interpreter_ui_delegate.report_status(status_text)
                logger.debug("Installer command: %s", command)

                if self.target_path and os.path.exists(self.target_path):
                    # Establish a CWD for the command, but remove it afterwards
                    # for safety. We'd better not rely on this, many tasks can be
                    # fiddling with the CWD at the same time.
                    def dispatch():
                        prev_cwd = os.getcwd()
                        os.chdir(self.target_path)
                        try:
                            return method(params)
                        finally:
                            os.chdir(prev_cwd)

                    AsyncCall(dispatch, self._iter_commands)
                else:
                    AsyncCall(method, self._iter_commands, params)
            else:
                logger.debug("Commands %d out of %s completed", self.current_command, len(commands))
                self._finish_install()
        except Exception as ex:
            # Redirect errors to the delegate, instead of the default ErrorDialog.
            logger.exception("Error during installation: %s", ex)
            self.interpreter_ui_delegate.report_error(ex)

    @staticmethod
    def _get_command_name_and_params(command_data):
        if isinstance(command_data, dict):
            command_name = list(command_data.keys())[0]
            command_params = command_data[command_name]
        else:
            command_name = command_data
            command_params = {}
        command_name = command_name.replace("-", "_")
        # Prevent private methods from being accessed as commands
        command_name = command_name.strip("_")
        return command_name, command_params

    def _map_command(self, command_data):
        """Map a directive from the `installer` section to an internal
        method."""
        command_name, command_params = self._get_command_name_and_params(command_data)
        if not hasattr(self, command_name):
            raise ScriptingError(_('The command "%s" does not exist.') % command_name)
        return getattr(self, command_name), command_params

    def _finish_install(self):
        game_id = self.installer.save()
        path = None

        if path and AUTO_EXE_PREFIX not in path and not os.path.isfile(path) and self.installer.runner != "web":
            status = (
                _(
                    "The executable at path %s can't be found, please check the destination folder.\n"
                    "Some parts of the installation process may have not completed successfully."
                )
                % path
            )
            logger.warning("No executable found at specified location %s", path)
        else:
            status = self.installer.script.get("install_complete_text") or _("Installation completed!")
        AsyncCall(download_lutris_media, None, self.installer.game_slug)
        self.interpreter_ui_delegate.report_finished(game_id, status)

    def cleanup(self):
        """Clean up install dir after a successful install"""
        os.chdir(os.path.expanduser("~"))
        system.delete_folder(self.cache_path)

    def revert(self, remove_game_dir=True, completion_function=None, error_function=None):
        """Revert installation in case of an error. Since winekill can be slow,
        this runs asynchronously and calls cocompletion_function() when successful,
        or error_function(err) if it fails."""
        logger.info("Cancelling installation of %s", self.installer.game_name)

        self.cancelled = True

        def on_complete(_result, error):
            if error:
                error_function(error)
                return

            try:
                if self.abort_current_task:
                    self.abort_current_task()

                if self.target_path and remove_game_dir:
                    system.remove_folder(self.target_path)

                completion_function()
            except Exception as ex:
                error_function(ex)

        if self.installer.runner.startswith("wine"):
            AsyncCall(self.task, on_complete, {"name": "winekill"})
        else:
            on_complete(None, None)

    def _get_string_replacements(self):
        """Return a mapping of variables to their actual value"""

        current_res = self.current_resolution

        replacements = {
            "GAMEDIR": self.target_path,
            "SCRIPTDIR": self.installer.scriptdir,
            "CACHE": self.cache_path,
            "HOME": os.path.expanduser("~"),
            "STEAM_DATA_DIR": steam.steam().steam_data_dir,
            "DISC": self.game_disc,
            "USER": os.getenv("USER"),
            "INPUT": self.user_inputs[-1]["value"] if self.user_inputs else "",
            "VERSION": self.installer.version,
            "RESOLUTION": "x".join(current_res),
            "RESOLUTION_WIDTH": current_res[0],
            "RESOLUTION_HEIGHT": current_res[1],
        }

        try:
            replacements["RESOLUTION_WIDTH_HEX"] = hex(int(current_res[0]))
            replacements["RESOLUTION_HEIGHT_HEX"] = hex(int(current_res[1]))
        except (ValueError, TypeError):
            pass  # If we can't generate hex, just omit the vars

        try:
            replacements["WINEBIN"] = self.get_wine_path()
        except MisconfigurationError:
            pass  # If we can't get the path, just omit it

        # None values stringify as 'None', which is not what you want, so we'll
        # remove then pre-emptively. This happens for game install scripts that have
        # no 'self.target_path'.

        for key in [key for key, value in replacements.items() if value is None]:
            del replacements[key]

        replacements.update(self.installer.variables)
        # Add 'INPUT_<id>' replacements for user inputs with an id
        for input_data in self.user_inputs:
            alias = input_data["alias"]
            if alias:
                replacements[alias] = input_data["value"]
        replacements.update(self.game_files)
        return replacements

    def _substitute(self, template_string):
        """Replace path aliases with real paths."""
        if template_string is None:
            logger.warning("No template string given")
            return ""
        if str(template_string).replace("-", "_") in self.game_files:
            template_string = template_string.replace("-", "_")
        return system.substitute(template_string, self._get_string_replacements())

    def eject_wine_disc(self):
        """Use Wine to eject a CD, otherwise Wine can have problems detecting disc changes"""
        wine_path = self.get_wine_path()
        wine.eject_disc(wine_path, self.target_path)



import json
import os
from gettext import gettext as _
from typing import List

from PIL import Image

from lutris import settings
from lutris.runners.dolphin import PLATFORMS
from lutris.services.base import BaseService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import system
from lutris.util.dolphin.cache_reader import DOLPHIN_GAME_CACHE_FILE, DolphinCacheReader
from lutris.util.strings import slugify


class DolphinBanner(ServiceMedia):
    service = "dolphin"
    source = "local"
    size = (96, 32)
    file_patterns = ["%s.png"]
    dest_path = os.path.join(settings.CACHE_DIR, "dolphin/banners/small")


class DolphinService(BaseService):
    id = "dolphin"
    icon = "dolphin"
    name = _("Dolphin")
    runner = "dolphin"
    local = True
    medias = {"icon": DolphinBanner}

    def load(self):
        if not system.path_exists(DOLPHIN_GAME_CACHE_FILE):
            return
        cache_reader = DolphinCacheReader()
        dolphin_games = [DolphinGame.new_from_cache(game) for game in cache_reader.get_games()]
        for game in dolphin_games:
            game.save()
        return dolphin_games

    def generate_installer(self, db_game):
        details = json.loads(db_game["details"])
        return {
            "name": db_game["name"],
            "version": "Dolphin",
            "slug": db_game["slug"],
            "game_slug": self.get_installed_slug(db_game),
            "runner": self.get_installed_runner_name(db_game),
            "script": {
                "game": {"main_file": details["path"], "platform": details["platform"]},
            },
        }

    def get_installed_runner_name(self, db_game):
        return self.runner

    def get_game_directory(self, installer):
        """Pull install location from installer"""
        return os.path.dirname(installer["script"]["game"]["main_file"])

    def get_game_platforms(self, db_game: dict) -> List[str]:
        details_json = db_game.get("details")
        if details_json:
            details = json.loads(details_json)
            if details and details.get("platform"):
                platform_value = details["platform"]
                if platform_value.isdigit():
                    platform_number = int(details["platform"])
                    if 0 <= platform_number < len(PLATFORMS):
                        platform = PLATFORMS[platform_number]
                        return [platform]

                return [platform_value]
        return []


class DolphinGame(ServiceGame):
    """Game for the Dolphin emulator"""

    service = "dolphin"
    runner = "dolphin"
    installer_slug = "dolphin"

    @classmethod
    def new_from_cache(cls, cache_entry):
        """Create a service game from an entry from the Dolphin cache"""
        name = cache_entry["internal_name"] or os.path.splitext(cache_entry["file_name"])[0]
        service_game = cls()
        service_game.name = name
        service_game.appid = str(cache_entry["game_id"])
        service_game.slug = slugify(name)
        service_game.icon = service_game.get_banner(cache_entry)

        service_game.details = json.dumps({"path": cache_entry["file_path"], "platform": cache_entry["platform"][:-1]})
        return service_game

    @staticmethod
    def get_game_name(cache_entry):
        names = cache_entry["long_names"]
        name_index = 1 if len(names.keys()) > 1 else 0
        return str(names[list(names.keys())[name_index]])

    def get_banner(self, cache_entry):
        banner = DolphinBanner()
        banner_path = banner.get_possible_media_paths(self.appid)[0].path  # Dolphin only supports one media type

        if os.path.exists(banner_path):
            return banner_path

        (width, height), data = cache_entry["volume_banner"]
        if data:
            img = Image.frombytes("RGB", (width, height), data, "raw", ("BGRX"))
            # 96x32 is a bit small, maybe 2x scale?
            # img.resize((width * 2, height * 2))
            img.save(banner_path)
            return banner_path

        return ""



"""TreeView based game list"""

from gettext import gettext as _

# Third Party Libraries
# pylint: disable=no-member
from gi.repository import Gdk, Gtk, Pango

# Lutris Modules
from lutris import settings
from lutris.gui.views import (
    COL_ID,
    COL_INSTALLED,
    COL_INSTALLED_AT,
    COL_INSTALLED_AT_TEXT,
    COL_LASTPLAYED,
    COL_LASTPLAYED_TEXT,
    COL_MEDIA_PATHS,
    COL_NAME,
    COL_PLATFORM,
    COL_PLAYTIME,
    COL_PLAYTIME_TEXT,
    COL_RUNNER_HUMAN_NAME,
    COL_SORTNAME,
    COL_YEAR,
    COLUMN_NAMES,
)
from lutris.gui.views.base import GameView
from lutris.gui.widgets.cellrenderers import GridViewCellRendererImage


class GameListView(Gtk.TreeView, GameView):  # type:ignore[misc]
    """Show the main list of games."""

    __gsignals__ = GameView.__gsignals__

    def __init__(self, store):
        Gtk.TreeView.__init__(self)
        GameView.__init__(self)

        self.set_rules_hint(True)

        # Image column
        if settings.SHOW_MEDIA:
            self.image_renderer = GridViewCellRendererImage()
            self.media_column = Gtk.TreeViewColumn(
                "", self.image_renderer, media_paths=COL_MEDIA_PATHS, is_installed=COL_INSTALLED, game_id=COL_ID
            )
            self.media_column.set_reorderable(True)
            self.media_column.set_sort_indicator(False)
            self.media_column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
            self.append_column(self.media_column)
        else:
            self.image_renderer = None
            self.media_column = None

        self.set_game_store(store)

        # Text columns
        default_text_cell = self.set_text_cell()
        name_cell = self.set_text_cell()
        name_cell.set_padding(5, 0)

        self.set_column(name_cell, _("Name"), COL_NAME, 200, always_visible=True, sort_id=COL_SORTNAME)
        self.set_column(default_text_cell, _("Year"), COL_YEAR, 60)
        self.set_column(default_text_cell, _("Runner"), COL_RUNNER_HUMAN_NAME, 120)
        self.set_column(default_text_cell, _("Platform"), COL_PLATFORM, 120)
        self.set_column(default_text_cell, _("Last Played"), COL_LASTPLAYED_TEXT, 120, sort_id=COL_LASTPLAYED)
        self.set_column(default_text_cell, _("Play Time"), COL_PLAYTIME_TEXT, 100, sort_id=COL_PLAYTIME)
        self.set_column(default_text_cell, _("Installed At"), COL_INSTALLED_AT_TEXT, 120, sort_id=COL_INSTALLED_AT)

        self.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        self.connect_signals()
        self.connect("row-activated", self.on_row_activated)
        self.get_selection().connect("changed", self.on_cursor_changed)

    def set_game_store(self, game_store):
        super().set_game_store(game_store)

        self.model = game_store.store
        self.set_model(self.model)

        if self.media_column:
            size = game_store.service_media.size
            media_width = size[0]
            self.media_column.set_fixed_width(media_width)

    @staticmethod
    def set_text_cell():
        text_cell = Gtk.CellRendererText()
        text_cell.set_padding(10, 0)
        text_cell.set_property("ellipsize", Pango.EllipsizeMode.END)
        return text_cell

    def set_column(self, cell, header, column_id, default_width, always_visible=False, sort_id=None):
        column = Gtk.TreeViewColumn(header, cell, markup=column_id)
        column.set_sort_indicator(True)
        column.set_sort_column_id(column_id if sort_id is None else sort_id)
        column.set_resizable(True)
        column.set_reorderable(True)
        width = settings.read_setting("%s_column_width" % COLUMN_NAMES[column_id], section="list view")
        is_visible = settings.read_setting("%s_visible" % COLUMN_NAMES[column_id], section="list view")
        column.set_fixed_width(int(width) if width else default_width)
        column.set_visible(is_visible == "True" or always_visible if is_visible else True)
        self.append_column(column)
        column.connect("notify::width", self.on_column_width_changed)
        column.get_button().connect("button-press-event", self.on_column_header_button_pressed)
        return column

    def get_path_at(self, x, y):
        path_at = self.get_path_at_pos(x, y)
        if path_at is None:
            return None
        path, _col, _cx, _cy = path_at
        return path

    def set_selected(self, paths, scroll_into_view=False):
        selection = self.get_selection()
        selection.unselect_all()

        for idx, path in enumerate(paths):
            selection.select_path(path)
            if scroll_into_view and idx == 0:
                self.scroll_to_cell(path, None, False, 0.0, 0.0)

    def get_selected(self):
        """Return list of all selected items"""
        selection = self.get_selection().get_selected_rows()
        if not selection:
            return None
        return selection[1]

    def get_game_id_for_path(self, path):
        iterator = self.get_model().get_iter(path)
        return self.get_model().get_value(iterator, COL_ID)

    def get_path_for_game_id(self, game_id):
        path_found = None

        def check_path(model, path, iterator):
            nonlocal path_found
            row_id = model.get_value(iterator, COL_ID)
            if game_id == row_id:
                path_found = path
                return True  # stop iteration

        self.get_model().foreach(check_path)
        return path_found

    def set_selected_game(self, game_id):
        row = self.game_store.get_row_by_id(game_id, filtered=True)
        if row:
            self.set_cursor(row.path)

    def on_column_header_button_pressed(self, button, event):
        """Handles column header button press events"""
        if event.button == Gdk.BUTTON_SECONDARY:
            menu = GameListColumnToggleMenu(self.get_columns())
            menu.popup_at_pointer(None)
            return True

    def on_row_activated(self, widget, line=None, column=None):
        """Handles double clicks"""
        selected_id = self.get_selected_game_id()
        if selected_id:
            self.emit("game-activated", selected_id)

    def on_cursor_changed(self, widget, _line=None, _column=None):
        selected_items = self.get_selected()
        self.emit("game-selected", selected_items)

    @staticmethod
    def on_column_width_changed(col, *args):
        col_name = col.get_title()
        if col_name:
            settings.write_setting(
                col_name.replace(" ", "") + "_column_width",
                col.get_fixed_width(),
                "list view",
            )


class GameListColumnToggleMenu(Gtk.Menu):
    def __init__(self, columns):
        super().__init__()
        self.columns = columns
        self.column_map = {}
        self.create_menuitems()
        self.show_all()

    def create_menuitems(self):
        for column in self.columns:
            title = column.get_title()
            if title == "":
                continue
            checkbox = Gtk.CheckMenuItem(title)
            checkbox.set_active(column.get_visible())
            if title == _("Name"):
                checkbox.set_sensitive(False)
            else:
                checkbox.connect("toggled", self.on_toggle_column)
            self.column_map[checkbox] = column
            self.append(checkbox)

    def on_toggle_column(self, check_menu_item):
        column = self.column_map[check_menu_item]
        is_visible = check_menu_item.get_active()
        column.set_visible(is_visible)
        settings.write_setting(
            column.get_title().replace(" ", "") + "_visible",
            str(is_visible),
            "list view",
        )



from gi.repository import GObject, Gtk

from lutris.gui.installer.file_box import InstallerFileBox
from lutris.util.log import logger


class InstallerFilesBox(Gtk.ListBox):
    """List box presenting all files needed for an installer"""

    max_downloads = 3

    __gsignals__ = {
        "files-ready": (GObject.SIGNAL_RUN_LAST, None, (bool,)),
        "files-available": (GObject.SIGNAL_RUN_LAST, None, ()),
    }

    def __init__(self):
        super().__init__()
        self.installer = None
        self.ready_files = set()
        self.available_files = set()
        self.installer_files_boxes = {}
        self._file_queue = []

    def load_installer(self, installer):
        self.stop_all()

        self.installer = installer
        self.available_files.clear()
        self.ready_files.clear()
        self.installer_files_boxes.clear()
        self._file_queue.clear()

        for child in self.get_children():
            child.destroy()

        for installer_file in installer.files:
            installer_file_box = InstallerFileBox(installer_file)
            installer_file_box.connect("file-ready", self.on_file_ready)
            installer_file_box.connect("file-unready", self.on_file_unready)
            installer_file_box.connect("file-available", self.on_file_available)
            self.installer_files_boxes[installer_file.id] = installer_file_box
            self.add(installer_file_box)
            if installer_file_box.is_ready:
                self.ready_files.add(installer_file.id)
        self.show_all()
        self.check_files_ready()

    def start_all(self):
        """Iterates through installer files while keeping the number
        of simultaneously downloaded files down to a maximum number"""

        if len(self.available_files) == len(self.installer.files):
            logger.info("All files remain available")
            self.emit("files-available")
            return

        started_downloads = 0
        for file_id, file_entry in self.installer_files_boxes.items():
            if file_id not in self.available_files:
                if file_entry.provider == "download":
                    started_downloads += 1
                    if started_downloads <= self.max_downloads:
                        file_entry.start()
                    else:
                        self._file_queue.append(file_id)
                else:
                    file_entry.start()

    def stop_all(self):
        """Stops all ongoing files gathering.
        Iterates through installer files, and call the "stop" command
        if they've been started and not available yet.
        """
        self._file_queue.clear()
        for file_id, file_box in self.installer_files_boxes.items():
            if file_box.started and file_id not in self.available_files and file_box.stop_func is not None:
                file_box.stop_func()

    @property
    def is_ready(self):
        """Return True if all files are ready to be fetched"""
        return len(self.ready_files) == len(self.installer.files)

    def check_files_ready(self):
        """Checks if all installer files are ready and emit a signal if so"""
        self.emit("files-ready", self.is_ready)

    def on_file_ready(self, widget):
        """Fired when a file has a valid provider.
        If the file is user provided, it must set to a valid path.
        """
        file_id = widget.installer_file.id
        self.ready_files.add(file_id)
        self.check_files_ready()

    def on_file_unready(self, widget):
        """Fired when a file can't be provided.
        Blocks the installer from continuing.
        """
        file_id = widget.installer_file.id
        self.ready_files.discard(file_id)
        self.check_files_ready()

    def on_file_available(self, widget):
        """A new file is available"""
        file_id = widget.installer_file.id
        logger.debug("%s is available", file_id)
        self.available_files.add(file_id)
        if self._file_queue:
            next_file_id = self._file_queue.pop()
            self.installer_files_boxes[next_file_id].start()
        if len(self.available_files) == len(self.installer.files):
            logger.info("All files available")
            self.emit("files-available")

    def get_game_files(self):
        """Return a mapping of the local files usable by the interpreter"""
        files = {}
        for installer_file in self.installer.files:
            files.update(installer_file.get_dest_files_by_id())
        return files



"""Runner for the PICO-8 fantasy console"""

import json
import math
import os
import shutil
from gettext import gettext as _
from time import sleep

from lutris import settings
from lutris.database.games import get_game_by_field
from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.downloader import Downloader
from lutris.util.log import logger
from lutris.util.strings import split_arguments


class pico8(Runner):
    description = _("Runs PICO-8 fantasy console cartridges")
    multiple_versions = False
    human_name = _("PICO-8")
    platforms = [_("PICO-8")]
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("Cartridge file/URL/ID"),
            "help": _("You can put a .p8.png file path, URL, or BBS cartridge ID here."),
        }
    ]

    runner_options = [
        {
            "option": "fullscreen",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Fullscreen"),
            "default": True,
            "help": _("Launch in fullscreen."),
        },
        {
            "option": "window_size",
            "section": _("Graphics"),
            "label": _("Window size"),
            "type": "string",
            "default": "640x512",
            "help": _("The initial size of the game window."),
        },
        {
            "option": "splore",
            "type": "bool",
            "label": _("Start in splore mode"),
            "default": False,
        },
        {
            "option": "args",
            "type": "string",
            "label": _("Extra arguments"),
            "default": "",
            "help": _("Extra arguments to the executable"),
            "advanced": True,
        },
        {
            "option": "engine",
            "type": "string",
            "label": _("Engine (web only)"),
            "default": "pico8_0111g_4",
            "help": _("Name of engine (will be downloaded) or local file path"),
        },
    ]

    system_options_override = [{"option": "disable_runtime", "default": True}]

    runner_executable = "pico8/web.py"

    def __init__(self, config=None):
        super().__init__(config)

    @property
    def is_native(self):
        return self.runner_config.get("runner_executable", "") != ""

    @property
    def engine_path(self):
        engine = self.runner_config.get("engine")
        if not engine.lower().endswith(".js") and not os.path.exists(engine):
            engine = os.path.join(
                settings.RUNNER_DIR,
                "pico8/web/engines",
                self.runner_config.get("engine") + ".js",
            )
        return engine

    @property
    def cart_path(self):
        main_file = self.game_config.get("main_file")
        if self.is_native and main_file.startswith("http"):
            return os.path.join(settings.RUNNER_DIR, "pico8/cartridges", "tmp.p8.png")
        if not os.path.exists(main_file) and main_file.isdigit():
            return os.path.join(settings.RUNNER_DIR, "pico8/cartridges", main_file + ".p8.png")
        return main_file

    @property
    def launch_args(self):
        if self.is_native:
            args = self.get_command()
            args.append("-windowed")
            args.append("0" if self.runner_config.get("fullscreen") else "1")
            if self.runner_config.get("splore"):
                args.append("-splore")

            size = self.runner_config.get("window_size").split("x")
            if len(size) == 2:
                args.append("-width")
                args.append(size[0])
                args.append("-height")
                args.append(size[1])
            extra_args = self.runner_config.get("args", "")
            for arg in split_arguments(extra_args):
                args.append(arg)
        else:
            args = self.get_command() + [
                os.path.join(settings.RUNNER_DIR, "pico8/web/player.html"),
                "--window-size",
                self.runner_config.get("window_size"),
            ]
        return args

    def get_run_data(self):
        return {"command": self.launch_args, "env": self.get_env(os_env=False)}

    def is_installed(self, flatpak_allowed: bool = True) -> bool:
        """Checks if pico8 runner is installed and if the pico8 executable available."""
        if self.is_native and system.path_exists(self.runner_config.get("runner_executable")):
            return True
        return system.path_exists(os.path.join(settings.RUNNER_DIR, "pico8/web/player.html"))

    def prelaunch(self):
        if not self.game_config.get("main_file") and self.is_installed():
            return True
        if os.path.exists(os.path.join(settings.RUNNER_DIR, "pico8/cartridges", "tmp.p8.png")):
            os.remove(os.path.join(settings.RUNNER_DIR, "pico8/cartridges", "tmp.p8.png"))

        # Don't download cartridge if using web backend and cart is url
        if self.is_native or not self.game_config.get("main_file").startswith("http"):
            if not os.path.exists(self.game_config.get("main_file")) and (
                self.game_config.get("main_file").isdigit() or self.game_config.get("main_file").startswith("http")
            ):
                if not self.game_config.get("main_file").startswith("http"):
                    pid = int(self.game_config.get("main_file"))
                    num = math.floor(pid / 10000)
                    downloadUrl = "https://www.lexaloffle.com/bbs/cposts/" + str(num) + "/" + str(pid) + ".p8.png"
                else:
                    downloadUrl = self.game_config.get("main_file")
                cartPath = self.cart_path
                system.create_folder(os.path.dirname(cartPath))

                downloadCompleted = False

                def on_downloaded_cart():
                    nonlocal downloadCompleted
                    # If we are offline we don't want an empty file to overwrite the cartridge
                    if dl.downloaded_size:
                        shutil.move(cartPath + ".download", cartPath)
                    else:
                        os.remove(cartPath + ".download")
                    downloadCompleted = True

                dl = Downloader(
                    downloadUrl,
                    cartPath + ".download",
                    True,
                    callback=on_downloaded_cart,
                )
                dl.start()

                # Wait for download to complete or continue if it exists (to work in offline mode)
                while not os.path.exists(cartPath):
                    if downloadCompleted or dl.state == Downloader.ERROR:
                        logger.error("Could not download cartridge from %s", downloadUrl)
                        return False
                    sleep(0.1)

        # Download js engine
        if not self.is_native and not os.path.exists(self.runner_config.get("engine")):
            enginePath = os.path.join(
                settings.RUNNER_DIR,
                "pico8/web/engines",
                self.runner_config.get("engine") + ".js",
            )
            if not os.path.exists(enginePath):
                downloadUrl = "https://www.lexaloffle.com/bbs/" + self.runner_config.get("engine") + ".js"
                system.create_folder(os.path.dirname(enginePath))
                dl = Downloader(downloadUrl, enginePath, True)
                dl.start()
                dl.join()

    def play(self):
        launch_info = {}
        launch_info["env"] = self.get_env(os_env=False)

        game_data = get_game_by_field(self.config.game_config_id, "configpath")

        command = self.launch_args

        if self.is_native:
            if not self.runner_config.get("splore"):
                command.append("-run")
            cartPath = self.cart_path
            if not os.path.exists(cartPath):
                raise MissingGameExecutableError(filename=cartPath)
            command.append(cartPath)

        else:
            command.append("--name")
            command.append(game_data.get("name") + " - PICO-8")

            # icon = datapath.get_icon_path(game_data.get("slug"))
            # if icon:
            #     command.append("--icon")
            #     command.append(icon)

            webargs = {
                "cartridge": self.cart_path,
                "engine": self.engine_path,
                "fullscreen": self.runner_config.get("fullscreen") is True,
            }
            command.append("--execjs")
            command.append("load_config(" + json.dumps(webargs) + ")")

        launch_info["command"] = command
        return launch_info



import math
import time
from itertools import chain

from lutris import settings
from lutris.database import sql
from lutris.util.log import logger
from lutris.util.strings import slugify

_SERVICE_CACHE = {}
_SERVICE_CACHE_ACCESSED = False  # Keep time of last access to have a self degrading cache


def get_games(searches=None, filters=None, excludes=None, sorts=None):
    return sql.filtered_query(
        settings.DB_PATH, "games", searches=searches, filters=filters, excludes=excludes, sorts=sorts
    )


def get_games_where(**conditions):
    """
    Query games table based on conditions

    Args:
        conditions (dict): named arguments with each field matches its desired value.
        Special values for field names can be used:
            <field>__lessthan will return rows where `field` is less than the value
            <field>__isnull will return rows where `field` is NULL if the value is True
            <field>__not will invert the condition using `!=` instead of `=`
            <field>__in will match rows for every value of `value`, which should be an iterable

    Returns:
        list: Rows matching the query

    """
    query = "select * from games"
    condition_fields = []
    condition_values = []
    for field, value in conditions.items():
        field, *extra_conditions = field.split("__")
        if extra_conditions:
            extra_condition = extra_conditions[0]
            if extra_condition == "lessthan":
                condition_fields.append("{} < ?".format(field))
                condition_values.append(value)
            if extra_condition == "isnull":
                condition_fields.append("{} is {} null".format(field, "" if value else "not"))
            if extra_condition == "not":
                condition_fields.append("{} != ?".format(field))
                condition_values.append(value)
            if extra_condition == "in":
                if not hasattr(value, "__iter__"):
                    raise ValueError("Value should be an iterable (%s given)" % value)
                if len(value) > 999:
                    raise ValueError("SQLite limited to a maximum of 999 parameters.")
                if value:
                    condition_fields.append("{} in ({})".format(field, ", ".join("?" * len(value)) or ""))
                    condition_values = list(chain(condition_values, value))
        else:
            condition_fields.append("{} = ?".format(field))
            condition_values.append(value)
    condition = " AND ".join(condition_fields)
    if condition:
        query = " WHERE ".join((query, condition))
    else:
        # Inspect and document why we should return
        # an empty list when no condition is present.
        return []
    return sql.db_query(settings.DB_PATH, query, tuple(condition_values))


def get_games_by_ids(game_ids):
    # sqlite limits the number of query parameters to 999, to
    # bypass that limitation, divide the query in chunks
    size = 999
    return list(
        chain.from_iterable(
            [
                get_games_where(id__in=list(game_ids)[page * size : page * size + size])
                for page in range(math.ceil(len(game_ids) / size))
            ]
        )
    )


def get_game_for_service(service, appid):
    if service == "lutris":
        return get_game_by_field(appid, field="slug")

    existing_games = get_games(filters={"service_id": appid, "service": service})
    if existing_games:
        return existing_games[0]


def get_all_installed_game_for_service(service):
    if service == "lutris":
        db_games = get_games(filters={"installed": 1})
        return {g["slug"]: g for g in db_games}

    db_games = get_games(filters={"service": service, "installed": 1})
    return {g["service_id"]: g for g in db_games}


def get_service_games(service):
    """Return the list of all installed games for a service"""
    global _SERVICE_CACHE_ACCESSED
    previous_cache_accessed = _SERVICE_CACHE_ACCESSED or 0
    _SERVICE_CACHE_ACCESSED = time.time()
    if service not in _SERVICE_CACHE or _SERVICE_CACHE_ACCESSED - previous_cache_accessed > 1:
        if service == "lutris":
            _SERVICE_CACHE[service] = [game["slug"] for game in get_games(filters={"installed": "1"})]
        else:
            _SERVICE_CACHE[service] = [
                game["service_id"] for game in get_games(filters={"service": service, "installed": "1"})
            ]
    return _SERVICE_CACHE[service]


def get_game_by_field(value, field="slug"):
    """Query a game based on a database field"""
    if field not in ("slug", "installer_slug", "id", "configpath", "name"):
        raise ValueError("Can't query by field '%s'" % field)
    game_result = sql.db_select(settings.DB_PATH, "games", condition=(field, value))
    if game_result:
        return game_result[0]
    return {}


def get_games_by_runner(runner):
    """Return all games using a specific runner"""
    return sql.db_select(settings.DB_PATH, "games", condition=("runner", runner))


def get_games_by_slug(slug):
    """Return all games using a specific slug"""
    return sql.db_select(settings.DB_PATH, "games", condition=("slug", slug))


def add_game(**game_data):
    """Add a game to the database."""
    game_data["installed_at"] = int(time.time())
    if "slug" not in game_data:
        game_data["slug"] = slugify(game_data["name"])
    return sql.db_insert(settings.DB_PATH, "games", game_data)


def add_games_bulk(games):
    """
    Add a list of games to the database.
    The dicts must have an identical set of keys.

    Args:
        games (list): list of games in dict format
    Returns:
        list: List of inserted game ids
    """
    return [sql.db_insert(settings.DB_PATH, "games", game) for game in games]


def add_or_update(**params):
    """Add a game to the database or update an existing one

    If an 'id' is provided in the parameters then it
    will try to match it, otherwise it will try matching
    by slug, creating one when possible.
    """
    game_id = update_existing(**params)
    if game_id:
        return game_id

    return add_game(**params)


def update_existing(**params):
    """Updates a game, but do not add one. If the game exists, this returns its ID;
    if not it returns None and makes no changes."""
    game_id = get_matching_game(params)
    if game_id:
        params["id"] = game_id
        sql.db_update(settings.DB_PATH, "games", params, {"id": game_id})
        return game_id
    return None


def get_matching_game(params):
    """Tries to match given parameters with an existing game"""
    # Always match by ID if provided
    if params.get("id"):
        game = get_game_by_field(params["id"], "id")
        if game:
            return game["id"]
        logger.warning("Game ID %s provided but couldn't be matched", params["id"])
    slug = params.get("slug") or slugify(params.get("name"))
    if not slug:
        raise ValueError("Can't add or update without an identifier")
    for game in get_games_by_slug(slug):
        if game["installed"]:
            if game["configpath"] == params.get("configpath"):
                return game["id"]
        else:
            if game["runner"] == params.get("runner") or not all([params.get("runner"), game["runner"]]):
                return game["id"]
    return None


def delete_game(game_id):
    """Delete a game from the PGA."""
    sql.db_delete(settings.DB_PATH, "games", "id", game_id)


def get_used_runners():
    """Return a list of the runners in use by installed games."""
    with sql.db_cursor(settings.DB_PATH) as cursor:
        query = "select distinct runner from games where runner is not null order by runner"
        rows = cursor.execute(query)
        results = rows.fetchall()
    return [result[0] for result in results if result[0]]


def get_used_platforms():
    """Return a list of platforms currently in use"""
    with sql.db_cursor(settings.DB_PATH) as cursor:
        query = (
            "select distinct platform from games where platform is not null and platform is not '' order by platform"
        )
        rows = cursor.execute(query)
        results = rows.fetchall()
    return [result[0] for result in results if result[0]]


def get_game_count(param, value):
    res = sql.db_select(settings.DB_PATH, "games", fields=("COUNT(id)",), condition=(param, value))
    if res:
        return res[0]["COUNT(id)"]



"""Set service ID for Steam games"""

from lutris import settings
from lutris.database.games import get_games, sql


def migrate():
    """Run migration"""
    for game in get_games():
        if not game.get("steamid"):
            continue
        if game["runner"] and game["runner"] != "steam":
            continue
        print("Migrating Steam game %s" % game["name"])
        sql.db_update(
            settings.DB_PATH, "games", {"service": "steam", "service_id": game["steamid"]}, {"id": game["id"]}
        )



import json
import os
from gettext import gettext as _
from typing import Any, Dict, Iterable, List, Optional

from gi.repository import Gio

from lutris import settings
from lutris.api import get_api_games, get_game_installers, read_api_key
from lutris.database.games import get_games
from lutris.database.services import ServiceGameCollection
from lutris.game import Game
from lutris.gui import dialogs
from lutris.gui.views.media_loader import download_media
from lutris.services.base import (
    SERVICE_LOGIN,
    LutrisBanner,
    LutrisCoverart,
    LutrisCoverartMedium,
    LutrisIcon,
    OnlineService,
)
from lutris.services.service_game import ServiceGame
from lutris.util import http
from lutris.util.busy import BusyAsyncCall
from lutris.util.log import logger


class LutrisGame(ServiceGame):
    """Service game created from the Lutris API"""

    service = "lutris"

    @classmethod
    def new_from_api(cls, api_payload):
        """Create an instance of LutrisGame from the API response"""
        service_game = LutrisGame()
        service_game.appid = api_payload["slug"]
        service_game.slug = api_payload["slug"]
        service_game.name = api_payload["name"]
        service_game.details = json.dumps(api_payload)
        return service_game


class LutrisService(OnlineService):
    """Service for Lutris games"""

    id = "lutris"
    name = _("Lutris")
    icon = "net.lutris.Lutris"
    online = True
    medias = {
        "icon": LutrisIcon,
        "banner": LutrisBanner,
        "coverart_med": LutrisCoverartMedium,
        "coverart_big": LutrisCoverart,
    }
    default_format = "coverart_med"

    api_url = settings.SITE_URL + "/api"
    login_url = settings.SITE_URL + "/api/accounts/token"
    cache_path = os.path.join(settings.CACHE_DIR, "lutris")
    token_path = os.path.join(settings.CACHE_DIR, "auth-token")

    @property
    def credential_files(self):
        """Return a list of all files used for authentication"""
        return [self.token_path]

    def match_games(self):
        """Matching lutris games is much simpler... No API call needed."""
        service_games = {str(game["appid"]): game for game in ServiceGameCollection.get_for_service(self.id)}
        for lutris_game in get_games():
            self.match_game(service_games.get(lutris_game["slug"]), lutris_game)

    def is_connected(self):
        """Is the service connected?"""
        return self.is_authenticated()

    def login(self, parent=None):
        """Connect to Lutris"""
        login_dialog = dialogs.ClientLoginDialog(parent=parent)
        login_dialog.connect("connected", self.on_connect_success)

    def on_connect_success(self, _widget, _username):
        """Handles connection success"""
        SERVICE_LOGIN.fire(self)

    def get_library(self):
        """Return the remote library as a list of dicts."""
        credentials = read_api_key()
        if not credentials:
            return []
        url = settings.SITE_URL + "/api/users/library"
        request = http.Request(url, headers={"Authorization": "Token " + credentials["token"]})
        try:
            response = request.get()
        except http.HTTPError as ex:
            logger.error("Unable to load library: %s", ex)
            return []
        return response.json

    def load(self):
        lutris_games = self.get_library()
        logger.debug("Loaded %s games from Lutris library", len(lutris_games))
        for game in lutris_games:
            lutris_game = LutrisGame.new_from_api(game)
            lutris_game.save()
        logger.debug("Matching with already installed games")
        self.match_games()
        logger.debug("Lutris games loaded")
        return lutris_games

    def load_icons(self):
        super().load_icons()
        # Also load any media for games that use Lutris media,
        # but are not in the Lutris library.
        sync_media()

    def get_installed_slug(self, db_game):
        return db_game["slug"]

    def install(self, db_game):
        slug = db_game["slug"]
        return self.install_by_id(slug)

    def install_by_id(self, appid):
        def on_installers_ready(installers, error):
            if error:
                raise error  # bounce any error off the backstop
            if not installers:
                raise RuntimeError(_("Lutris has no installers for %s. Try using a different service instead.") % appid)
            application = Gio.Application.get_default()
            application.show_installer_window(installers)

        BusyAsyncCall(get_game_installers, on_installers_ready, appid)  # appid is the slug for Lutris games

    def get_installed_runner_name(self, db_game):
        platforms = self.get_game_platforms(db_game)

        if platforms and len(platforms) == 1:
            platform = platforms[0].casefold()

            if platform == "windows":
                return "wine"

            if platform == "linux":
                return "linux"

            if platform == "ms-dos":
                return "dosbox"

        return ""

    def get_game_platforms(self, db_game: dict) -> List[str]:
        details = db_game.get("details")
        if details:
            platforms = json.loads(details).get("platforms")
            if platforms is not None:
                return [p.get("name") for p in platforms]
        return []

    def get_service_db_game(self, game: Game):
        if game.service == self.id and game.slug:
            return ServiceGameCollection.get_game(self.id, game.slug)
        return None


def download_lutris_media(slug):
    """Download all media types for a single lutris game"""
    url = settings.SITE_URL + "/api/games/%s" % slug
    request = http.Request(url)
    try:
        response = request.get()
    except http.HTTPError as ex:
        logger.debug("Unable to load %s: %s", slug, ex)
        return
    response_data = response.json
    icon_url = _get_response_game_icon(response_data)
    if icon_url:
        download_media({slug: icon_url}, LutrisIcon())

    banner_url = _get_response_game_banner(response_data)
    if banner_url:
        download_media({slug: banner_url}, LutrisBanner())

    coverart_url = _get_response_game_coverart(response_data)
    if coverart_url:
        download_media({slug: coverart_url}, LutrisCoverart())


def sync_media(slugs: Optional[Iterable[str]] = None) -> Dict[str, int]:
    """Download missing media for Lutris games; if a set of slugs
    is not provided, downloads them for all games in the PGA."""
    if slugs is None:
        slugs = {game["slug"] for game in get_games()}
    else:
        slugs = set(s for s in slugs if s)

    if not slugs:
        return {}

    banners_available = {fn.split(".")[0] for fn in os.listdir(settings.BANNER_PATH)}
    icons_available = {
        fn.split(".")[0].replace("lutris_", "") for fn in os.listdir(settings.ICON_PATH) if fn.startswith("lutris_")
    }
    covers_available = {fn.split(".")[0] for fn in os.listdir(settings.COVERART_PATH)}
    complete_games = banners_available.intersection(icons_available).intersection(covers_available)

    slugs_to_download = slugs - complete_games
    if not slugs_to_download:
        return {}
    games = get_api_games(list(slugs_to_download))

    alias_map = {}
    api_slugs = set()
    for game in games:
        api_slugs.add(game["slug"])
        for alias in game["aliases"]:
            if alias["slug"] in slugs_to_download:
                alias_map[game["slug"]] = alias["slug"]
    alias_slugs = set(alias_map.values())
    used_alias_slugs = alias_slugs - api_slugs
    for alias_slug in used_alias_slugs:
        for game in games:
            if alias_slug in [alias["slug"] for alias in game["aliases"]]:
                game["slug"] = alias_map[game["slug"]]
                continue
    banner_urls = {
        game["slug"]: _get_response_game_banner(game)
        for game in games
        if game["slug"] not in banners_available and _get_response_game_banner(game)
    }
    icon_urls = {
        game["slug"]: _get_response_game_icon(game)
        for game in games
        if game["slug"] not in icons_available and _get_response_game_icon(game)
    }
    coverart_urls = {
        game["slug"]: _get_response_game_coverart(game)
        for game in games
        if game["slug"] not in covers_available and _get_response_game_coverart(game)
    }
    logger.debug("Syncing %s banners, %s icons and %s covers", len(banner_urls), len(icon_urls), len(coverart_urls))
    downloaded_banners = download_media(banner_urls, LutrisBanner())
    downloaded_icons = download_media(icon_urls, LutrisIcon())
    downloaded_covers = download_media(coverart_urls, LutrisCoverart())
    return {
        "banners": len(downloaded_banners),
        "icons": len(downloaded_icons),
        "covers": len(downloaded_covers),
    }


def _get_response_game_coverart(api_game: Dict[str, Any]) -> Optional[str]:
    return api_game.get("coverart")


def _get_response_game_banner(api_game: Dict[str, Any]) -> Optional[str]:
    return api_game.get("banner_url") or api_game.get("banner")


def _get_response_game_icon(api_game: Dict[str, Any]) -> Optional[str]:
    return api_game.get("icon_url") or api_game.get("icon")
